import __main__ as _main_module
import datetime
import glob
import inspect
import os
from io import BytesIO

import dill
import matplotlib.colors as Colors
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import win32clipboard as clip
import win32con
from IPython import get_ipython
from PIL import ImageGrab
from matplotlib.collections import PolyCollection
from matplotlib.collections import QuadMesh
from scipy import optimize
from scipy.spatial import Voronoi
from scipy.special import erf

ipython = get_ipython()


class InteractivePlotLib:
    def __init__(self, doc_path=""):
        fig_numbers = plt.get_fignums()
        for i in fig_numbers:
            plt.close(plt.figure(i))
        self.figs = {}
        if doc_path:
            self.doc = Document(doc_path)
        else:
            self.doc = ""

    def garbage_collector(self):
        fig_numbers = plt.get_fignums()
        mark_to_delete = []
        for i in self.figs:
            if self.figs[i].fig.number not in fig_numbers:
                mark_to_delete.append(i)
        for i in mark_to_delete:
            del self.figs[i].fig
            del self.figs[i]

        for i in list(set(plt.get_fignums()) - set(self.figs.keys())):
            fig = plt.figure(i)
            self.figs[fig.number] = InteractivePlotLibFigure(fig, self.doc, self)

    def figure(self, ind=-1):
        self.garbage_collector()
        if ind in list(self.figs.keys()):
            self.figs[ind].fig.canvas.mpl_disconnect(self.figs[ind].cid)
            self.figs[ind].fig.canvas.mpl_disconnect(self.figs[ind].cid2)
            self.figs[ind].fig.canvas.mpl_disconnect(self.figs[ind].cid3)
            self.figs[ind].fig.canvas.mpl_disconnect(self.figs[ind].cid4)
        if ind < 0:
            fig = plt.figure()
        else:
            fig = plt.figure(ind)
        if len(fig.axes) == 0:
            plt.axes()
        self.figs[fig.number] = InteractivePlotLibFigure(fig, self.doc, self)
        return fig

    def load(self, number):
        pre_load_fig_numbers = plt.get_fignums()
        out = self.doc.load(number)
        self.garbage_collector()
        post_load_fig_numbers = plt.get_fignums()
        for i in list(set(post_load_fig_numbers) - set(pre_load_fig_numbers)):
            h = self.figure(i)
            for t in h.axes[0].texts:
                if len(t.get_text()) > 5 and t.get_text()[:5] == "Scan ":
                    t.remove()
            plt.show()
            plt.pause(1e-6)
        return out


def extract_data_from_mesh(obj):
    array = obj.get_array()
    width = obj._meshWidth
    height = obj._meshHeight
    data = np.reshape(array, [height, width])
    x = np.unique(obj._coordinates[:, :, 0].flatten())
    y = np.unique(obj._coordinates[:, :, 1].flatten())
    x2 = (x[:-1] + x[1:]) / 2
    y2 = (y[:-1] + y[1:]) / 2
    return data, x, y, x2, y2


def extract_data_from_pcolor(obj):
    array = obj.get_array()
    x_ext = np.unique(np.array([[vertex[0] for vertex in path.vertices] for path in obj._paths]).flatten())
    y_ext = np.unique(np.array([[vertex[1] for vertex in path.vertices] for path in obj._paths]).flatten())
    x_ext2 = (x_ext[:-1] + x_ext[1:]) / 2
    if len(y_ext) > 1:
        y_ext2 = (y_ext[:-1] + y_ext[1:]) / 2
    else:
        y_ext2 = y_ext
    x2d = np.tile(x_ext2, [len(y_ext2), 1])
    y2d = np.tile(y_ext2, [len(x_ext2), 1]).T
    print(len(y_ext2))
    data = np.reshape(array, [len(y_ext2), len(x_ext2)])
    return data, x2d, y2d, x_ext2, y_ext2, x_ext, y_ext


class InteractivePlotLibFigure:
    def __init__(self, fig, doc, master_obj):

        self.fig = fig
        self.ax = self.fig.axes[0]
        self.doc = doc
        self.master_obj = master_obj

        self.line_selected = None
        self.th = 0.2
        self.th_point_advantage = 2

        self.correct_order = [0]

        self.marker_mode = False
        self.marker_list = [[]]

        self.voronoi_mode = False
        self.voronoi_obj = InteractivePlotLibFigure.Voronoi_Qplt(self)

        self.grid_state = False

        self.fig.canvas.mpl_disconnect(self.fig.canvas.manager.key_press_handler_id)
        self.cid = self.fig.canvas.mpl_connect("button_press_event", self.mouse_click)
        self.cid2 = self.fig.canvas.mpl_connect("key_press_event", self.keyboard_click)
        self.cid3 = self.fig.canvas.mpl_connect("key_release_event", self.keyboard_release)
        self.cid4 = self.fig.canvas.mpl_connect("button_release_event", self.mouse_release)

        self.lines = []
        self.rectangle = None
        self.state = self.ProtoStateMachine(self, done=True)
        self.plot_type = "plot"

    def user_interaction(self, type, event):
        self.refresh_axis()
        if self.state.done:
            if type == "mouse_click":
                xlim = self.ax.get_xlim()
                ylim = self.ax.get_ylim()
                bbox = self.ax.get_window_extent().transformed(self.fig.dpi_scale_trans.inverted())  # axes bounding box

                if event.y < bbox.ymin * self.fig.dpi:
                    if event.x < (bbox.xmin + (bbox.xmax - bbox.xmin) / 5) * self.fig.dpi:
                        self.state = InteractivePlotLibFigure.StateMachineLim(self, "xstart")
                    elif event.x > (bbox.xmax - (bbox.xmax - bbox.xmin) / 5) * self.fig.dpi:
                        self.state = InteractivePlotLibFigure.StateMachineLim(self, "xend")
                    else:
                        self.state = InteractivePlotLibFigure.StateMachineLabel(self, "xlabel")

                elif event.y > bbox.ymax * self.fig.dpi:
                    self.state = InteractivePlotLibFigure.StateMachineLabel(self, "title")

                elif event.x < bbox.xmin * self.fig.dpi:
                    if event.y < (bbox.ymin + (bbox.ymax - bbox.ymin) / 5) * self.fig.dpi:
                        self.state = InteractivePlotLibFigure.StateMachineLim(self, "ystart")
                    elif event.y > (bbox.ymax - (bbox.ymax - bbox.ymin) / 5) * self.fig.dpi:
                        self.state = InteractivePlotLibFigure.StateMachineLim(self, "yend")
                    else:
                        self.state = InteractivePlotLibFigure.StateMachineLabel(self, "ylabel")

                elif event.x > bbox.xmax * self.fig.dpi:
                    self.state = InteractivePlotLibFigure.StateMachineLegend(self)

                else:
                    self.plot_type = "plot"
                    self.detect_curve_click(
                        event.xdata,
                        event.ydata,
                        xlim,
                        ylim,
                        bbox.width,
                        bbox.height,
                    )
                    if (
                        self.line_selected is not None
                        and event.button == 3
                        and hasattr(self.line_selected.obj, "InteractivePlotLib_Type")
                        and self.line_selected.obj.InteractivePlotLib_Type == "Fit"
                    ):
                        func = self.line_selected.obj.fit_func
                        xlim = self.ax.get_xlim()
                        x = np.linspace(min(xlim), max(xlim), 1001)
                        y = func(x)
                        self.line_selected.obj.set_xdata(x)
                        self.line_selected.obj.set_ydata(y)

                        print("FITTTTT")
                    if self.line_selected is None and isinstance(self.ax.get_children()[0], PolyCollection):
                        self.plot_type = "pcolor"
                        self.line_selected = InteractivePlotLibFigure.PcolorSelected(self, self.ax.get_children()[0])

                    if self.line_selected is None and isinstance(self.ax.get_children()[0], QuadMesh):
                        self.plot_type = "mesh"
                        self.line_selected = InteractivePlotLibFigure.MeshSelected(self, self.ax.get_children()[0])

            elif type == "keyboard_click":
                if event.key == "ctrl+v":
                    clip.OpenClipboard()
                    data = clip.GetClipboardData()
                    clip.CloseClipboard()
                    lines = data.split("\r\n")[:-1]
                    if len(lines[0].split("\t")) <= 2:
                        out = np.array(([i.split("\t") for i in data.split("\r\n")[:-1]]), "double")
                        if np.shape(out)[1] == 2:
                            self.ax.plot(out[:, 0], out[:, 1], linewidth=1)
                        elif np.shape(out)[1] == 1:
                            self.ax.plot(out[:, 1], linewidth=1)

                    elif (len(lines[0].split("\t")) > 2) and len(lines) == 2:
                        self.ax.plot(
                            np.array(lines[0].split("\t")[1:], "double"),
                            np.array(lines[1].split("\t")[1:], "double"),
                            linewidth=1,
                        )
                        self.ax.legend([np.array(lines[1].split("\t")[0], "double")])
                    else:
                        x_values = np.array(lines[0].split("\t")[1:], "double")
                        y_values = np.array([i.split("\t")[0] for i in lines[1:]], "double")
                        out_2d = np.array(([i.split("\t")[1:] for i in lines[1:]]), "double")
                        print(x_values)
                        print(y_values)
                        print(out_2d)
                        self.ax.pcolormesh(x_values, y_values, out_2d, shading="auto")

                if event.key == "m":
                    self.marker_mode = not self.marker_mode

                if event.key == "shift+m" or event.key == "M":
                    self.marker_list.append([])

                if event.key == "alt+m":
                    self.marker_list = [[]]
                    self.marker_mode = False

                if event.key == "a":
                    self.ax.axis("equal")

                if event.key == "v":
                    self.state = InteractivePlotLibFigure.StateMachineVoronoi(self)

                if event.key == "alt+v":
                    self.voronoi_obj.remove()

                if event.key == "f":
                    self.state = InteractivePlotLibFigure.StateMachineFit(self)

                if event.key == "g":
                    self.grid_state = not self.grid_state
                    self.ax.grid(self.grid_state)

                if event.key == "alt+f":
                    self.remove_fit()

                if event.key == "shift+s" or event.key == "S":
                    if self.doc.doc:
                        self.doc.doc(list(self.master_obj.figs.keys()))

                if event.key == "s":
                    if self.doc.doc:
                        self.doc.doc([plt.gcf().number])

                if event.key == "r":
                    self.state = InteractivePlotLibFigure.StateMachineRect(self)

                if event.key == "alt+r":
                    self.rectangle = None

                if event.key == "n":
                    self.master_obj.figure()

                if event.key == "alt+l":
                    self.ax.get_legend().set_visible(not self.ax.get_legend().get_visible())

                if event.key == "p" and self.plot_type == "plot":
                    self.convert_to_pcolormesh()
                    self.plot_type = "mesh"

                if self.line_selected:
                    if event.key == "t":
                        self.line_selected.transpose()

                    if event.key == "l" and (self.plot_type == "pcolor" or self.plot_type == "mesh"):
                        self.line_selected.convert_to_lines()
                        self.line_selected = None
                        self.plot_type = "plot"

                    if event.key == "c":
                        self.state = InteractivePlotLibFigure.StateMachineColor(self)

                    if event.key == ":":
                        self.state = InteractivePlotLibFigure.StateMachineCommand(self)

                    if event.key == "ctrl+c":
                        self.line_selected.copy_to_clipboard()

                    if event.key == "delete":
                        self.line_selected.delete()

                    if event.key == "alt+delete":
                        self.line_selected.delete_neg()

                    if event.key == "up":
                        self.line_selected.correct_order(1)

                    if event.key == "down":
                        self.line_selected.correct_order(-1)

                    if event.key.isnumeric():
                        self.line_selected.line_width = int(event.key)
                        self.line_selected.emphasize_line_width()

            elif type == "keyboard_release":
                pass  # {"key": event.key}
        else:
            self.state.event(type, event)

        if self.state.done:
            self.state = self.ProtoStateMachine(self, done=True)

        self.refresh_figure()

    def convert_to_pcolormesh(self):
        self.lines = self.get_lines()
        y_bare = [float(value.get_text()) for value in self.ax.get_legend().texts]
        xmat = []
        ymat = []
        datamat = []
        for i, line in enumerate(self.lines):
            xmat.append(line.get_xdata())
            ymat.append([y_bare[i]] * len(line.get_xdata()))
            datamat.append(line.get_ydata())
        for line in self.lines:
            line.remove()
        self.lines = []
        obj = self.ax.pcolormesh(xmat, ymat, datamat, shading="auto")
        _, x, y, _, _ = extract_data_from_mesh(obj)
        self.ax.set_xlim([min(x), max(x)])
        self.ax.set_ylim([min(y), max(y)])
        self.ax.get_legend().remove()

    def detect_curve_click(self, base_x, base_y, base_xlim, base_ylim, width, height):
        scale_convert = InteractivePlotLibFigure.ConvertLogLin([self.ax.get_xscale(), self.ax.get_yscale()])
        line_list = self.get_lines_fit_extended()
        candidate = {"line_index": -1, "distance_2": np.inf, "xy": (0, 0)}
        x0 = scale_convert.convert[0](base_x)
        y0 = scale_convert.convert[1](base_y)
        xlim = scale_convert.convert[0](base_xlim)
        ylim = scale_convert.convert[1](base_ylim)

        xdiff = np.sqrt((xlim[1] - xlim[0]) ** 2 / width**2)
        ydiff = np.sqrt((ylim[1] - ylim[0]) ** 2 / height**2)

        for j, line in enumerate(line_list):
            x = scale_convert.bound[0](line.get_xdata())
            y = scale_convert.bound[1](line.get_ydata())
            idx = (x >= base_xlim[0]) & (x <= base_xlim[1]) & (y >= base_ylim[0]) & (y <= base_ylim[1])
            idx = np.where(idx)[0]
            idx = list(set.union(set(idx), set(idx + 1), set(idx - 1)))
            idx = [i for i in idx if i >= 0 and i < len(x)]

            if len(idx) == 0:
                continue

            x = scale_convert.convert[0](x[idx])
            y = scale_convert.convert[1](y[idx])

            x_n = x / xdiff
            y_n = y / ydiff
            x0_n = x0 / xdiff
            y0_n = y0 / ydiff
            if line.get_linestyle() == "None":
                norm_2 = (
                    (x0_n - x_n) ** 2 + (y0_n - y_n) ** 2
                ) / self.th_point_advantage  # give advantage for choosing a point over a line
                ind = np.argmin(norm_2)
                if norm_2[ind] < candidate["distance_2"]:
                    candidate["line_index"] = j
                    candidate["distance_2"] = norm_2[ind]

                    candidate["xy"] = (
                        scale_convert.un_convert[0](x_n[ind] * xdiff),
                        scale_convert.un_convert[1](y_n[ind] * ydiff),
                    )
            else:
                for i in range(len(x_n) - 1):
                    d_2, x4, y4 = self.p4((x_n[i], y_n[i]), (x_n[i + 1], y_n[i + 1]), (x0_n, y0_n))

                    if d_2 < candidate["distance_2"]:
                        candidate["line_index"] = j
                        candidate["distance_2"] = d_2
                        candidate["xy"] = (
                            scale_convert.un_convert[0](x4 * xdiff),
                            scale_convert.un_convert[1](y4 * ydiff),
                        )

        self.line_selected = None  # deselect line if selected
        if candidate["distance_2"] < self.th**2:
            self.line_selected = InteractivePlotLibFigure.LineSelected(
                self,
                line_list[candidate["line_index"]],
                candidate["line_index"],
                candidate["xy"],
            )

    def remove_fit(self):
        if hasattr(self.ax, "fit_markers"):
            for marker in self.ax.fit_markers:
                marker.remove()
        lines = self.ax.get_lines()
        for line in lines:
            if hasattr(line, "InteractivePlotLib_Type") and (
                line.InteractivePlotLib_Type == "Fit" or line.InteractivePlotLib_Type == "Fit_markers"
            ):
                line.remove()

    def predefined_fit_pcolor(self, req="n", Type="pcolor"):
        if Type == "pcolor":
            data, x2d, y2d, x_ext2, y_ext2, x_ext, y_ext = extract_data_from_pcolor(self.line_selected.obj)
        else:
            data, x2d, y2d, x_ext2, y_ext2 = extract_data_from_mesh(self.line_selected.obj)

        fit_line_x = []
        fit_line_y = []
        if req == "n":
            data = -data
        if req == "p":
            pass
        for i, d in enumerate(data):
            fit_line_x.append(x_ext2[np.argmax(d)])
            fit_line_y.append([y_ext2[i]])
        (line,) = self.ax.plot(fit_line_x, fit_line_y, ".m")
        setattr(line, "InteractivePlotLib_Type", "Fit")

    def predefined_fit(self, req="g"):
        self.remove_fit()
        print(req)
        if req == "remove":
            return 0
        if req.isnumeric():
            Fit(int(req), self.ax)
        elif req == "g":
            Fit(lambda x: np.exp(-(x**2) / 2), self.ax)
        elif req == "e":
            Fit(lambda x: np.exp(x), self.ax)
        elif req == "l":
            Fit(lambda x: 1 / (1 + x**2), self.ax)
        elif req == "r":
            Fit(lambda x: erf(x), self.ax)
        elif req == "s":
            x = self.ax.get_lines()[0].get_xdata()
            y = self.ax.get_lines()[0].get_ydata()
            w = np.fft.fft(y)
            freqs = np.fft.fftfreq(len(x))
            new_f = freqs[1 : len(freqs // 2)]
            out_freq = new_f[np.argmax(np.abs(w[1 : len(freqs // 2)]))]
            omega = out_freq / np.mean(np.diff(x)) * 2 * np.pi
            xlim = self.ax.get_xlim()
            request_x_scale = 1 / omega / (xlim[1] - xlim[0])
            print(request_x_scale)
            Fit(lambda x: np.sin(x), self.ax, {"x_scale": request_x_scale})
        elif req == "S":
            ax = self.ax
            x = ax.get_lines()[0].get_xdata()
            y = ax.get_lines()[0].get_ydata()
            w = np.fft.fft(y)
            freqs = np.fft.fftfreq(len(x))
            new_w = w[1 : len(freqs // 2)]
            new_f = freqs[1 : len(freqs // 2)]

            ind = new_f > 0
            new_f = new_f[ind]
            new_w = new_w[ind]

            out_freq = new_f[np.argmax(np.abs(new_w))]
            omega = out_freq / np.mean(np.diff(x)) * 2 * np.pi
            xlim = ax.get_xlim()
            request_x_scale = 1 / omega / (xlim[1] - xlim[0])
            print(request_x_scale)
            y_extra_scale = np.diff(ax.get_ylim())[0]
            Fit(
                lambda x, a: a[3] + a[2] * np.sin(a[0] * x) * np.exp(-x * a[1]),
                ax,
                {
                    "initial_guess": [omega, 1, 1, 0],
                    "func": "sin*exp",
                    "y_scale": y_extra_scale,
                    "text": lambda a: f"f = {a[0] / (2 * np.pi)} [Hz]\n tau = {1 / a[1]} [s]\n amp = {a[2] * y_extra_scale}\n y_offset={a[3] * y_extra_scale}",
                },
            )
        else:
            idx = req.find(",")
            convert_function = eval("lambda x,a: " + req[:idx])
            Fit(
                convert_function,
                self.ax,
                {
                    "initial_guess": eval(req[(idx + 1) :]),
                    "func": "lambda x,a: " + req[:idx],
                },
            )

    def p4(self, p1, p2, p3):
        x1, y1 = p1
        x2, y2 = p2
        x3, y3 = p3
        dx, dy = x2 - x1, y2 - y1
        det = dx * dx + dy * dy
        if det == 0:
            return np.inf, 0, 0
        a = (dy * (y3 - y1) + dx * (x3 - x1)) / det
        if a < 0 or a > 1:
            return np.inf, 0, 0

        x4, y4 = x1 + a * dx, y1 + a * dy
        d_2 = (x4 - x3) ** 2 + (y4 - y3) ** 2
        return d_2, x4, y4

    class proto_2d_graphics:
        def delete(self_2d):
            if self_2d.sup_self.rectangle:
                x_list = self_2d.x
                y_list = self_2d.y
                data = self_2d.data
                ind_x = np.array([True] * len(x_list))
                ind_y = np.array([True] * len(y_list))

                ind_x = (x_list >= self_2d.sup_self.rectangle.x[0]) & (x_list <= self_2d.sup_self.rectangle.x[1])
                ind_y = (y_list >= self_2d.sup_self.rectangle.y[0]) & (y_list <= self_2d.sup_self.rectangle.y[1])
                self_2d.sup_self.rectangle = None
                x_list = x_list[ind_x]
                y_list = y_list[ind_y]
                data = np.array(data, "double")
                data[
                    np.where(ind_y)[0][0] : np.where(ind_y)[0][-1] + 1,
                    np.where(ind_x)[0][0] : np.where(ind_x)[0][-1] + 1,
                ] = np.NaN
                self_2d.obj.set_array(data.flatten())
                self_2d.data = data

            else:
                self_2d.obj.remove()
                self_2d.alive = False

        def transpose(self_2d):
            self_2d.obj.remove()
            xlim = self_2d.sup_self.ax.get_xlim()
            ylim = self_2d.sup_self.ax.get_ylim()
            x_new = self_2d.y
            y_new = self_2d.x
            data_new = self_2d.data.T
            print("HI")
            self_2d.obj = self_2d.plot(x_new, y_new, data_new)

            self_2d.x = x_new
            self_2d.y = y_new
            self_2d.data = data_new
            self_2d.sup_self.ax.set_xlim(ylim)
            self_2d.sup_self.ax.set_ylim(xlim)

        def copy_to_clipboard(self_2d):
            data = ""
            x_list = self_2d.x2
            y_list = self_2d.y2
            data = self_2d.data
            ind_x = np.array([True] * len(x_list))
            ind_y = np.array([True] * len(y_list))
            data_s = ""
            if self_2d.sup_self.rectangle:
                ind_x = (x_list >= self_2d.sup_self.rectangle.x[0]) & (x_list <= self_2d.sup_self.rectangle.x[1])
                ind_y = (y_list >= self_2d.sup_self.rectangle.y[0]) & (y_list <= self_2d.sup_self.rectangle.y[1])
                self_2d.sup_self.rectangle = None
                x_list = x_list[ind_x]
                y_list = y_list[ind_y]
                data = data[
                    np.where(ind_y)[0][0] : np.where(ind_y)[0][-1] + 1,
                    np.where(ind_x)[0][0] : np.where(ind_x)[0][-1] + 1,
                ]
                print(data)

            data_s += f" ,{','.join([str(i) for i in x_list])}\r\n".replace(",", "\t")
            for j in range(len(y_list)):
                data_s += f"{y_list[j]},{','.join([str(i) for i in data[j]])}\r\n".replace(",", "\t")
            clip.OpenClipboard()
            clip.EmptyClipboard()
            clip.SetClipboardData(win32con.CF_UNICODETEXT, data_s)
            clip.CloseClipboard()

        def convert_to_lines(self_2d):
            print("HI")
            self_2d.obj.remove()
            cmap = plt.cm.get_cmap("jet")
            if len(self_2d.y2) == 1:
                self_2d.sup_self.ax.plot(self_2d.x2, self_2d.data[0])
            else:
                for i in range(len(self_2d.y2)):
                    self_2d.sup_self.ax.plot(
                        self_2d.x2,
                        self_2d.data[i],
                        color=cmap(i / (len(self_2d.y2) - 1)),
                    )
            self_2d.sup_self.ax.legend(self_2d.y2)
            self_2d.sup_self.ax.set_xlim([min(self_2d.x2), max(self_2d.x2)])
            self_2d.sup_self.ax.set_ylim(
                [
                    min(np.array(self_2d.data).flatten()),
                    max(np.array(self_2d.data).flatten()),
                ]
            )

    class Voronoi_Qplt:
        def __init__(self_voronoi, sup_self):
            self_voronoi.sup_self = sup_self
            self_voronoi.markers = []
            self_voronoi.graphical_objects = []

        def remove(self_voronoi):
            for obj in self_voronoi.graphical_objects:
                obj.remove()
            self_voronoi.graphical_objects = []
            self_voronoi.markers = []

        def generate_voronoi(self_voronoi):
            xlim = self_voronoi.sup_self.ax.get_xlim()
            ylim = self_voronoi.sup_self.ax.get_ylim()
            sup_self = self_voronoi.sup_self

            for obj in self_voronoi.graphical_objects:
                obj.remove()
            self_voronoi.graphical_objects = []

            points = np.array(self_voronoi.markers)
            (l,) = sup_self.ax.plot(points[:, 0], points[:, 1], "om")
            self_voronoi.graphical_objects.append(l)

            if len(points) == 2:
                if points[0][1] - points[1][1] == 0:
                    x = (points[0][0] + points[1][0]) / 2
                    (l,) = sup_self.ax.plot([x, x], [max(ylim), min(ylim)], "m")
                elif points[0][0] - points[1][0] == 0:
                    y = (points[0][1] + points[1][1]) / 2
                    (l,) = sup_self.ax.plot([max(xlim), min(xlim)], [y, y], "m")
                else:
                    slope = (points[0][1] - points[1][1]) / (points[0][0] - points[1][0])
                    prep_slope = -1 / slope
                    point0 = np.array([points[0][0] + points[1][0], points[0][1] + points[1][1]]) / 2
                    xy = line_in_lim(point0, prep_slope, xlim, ylim)
                    (l,) = sup_self.ax.plot([xy[0][0], xy[1][0]], [xy[0][1], xy[1][1]], "m")
                    self_voronoi.graphical_objects.append(l)

            if len(points) > 2:
                vor = Voronoi(points)

                (l,) = sup_self.ax.plot(vor.vertices[:, 0], vor.vertices[:, 1], "*y")
                self_voronoi.graphical_objects.append(l)

                for simplex in vor.ridge_vertices:
                    simplex = np.asarray(simplex)
                    if np.all(simplex >= 0):
                        (l,) = sup_self.ax.plot(vor.vertices[simplex, 0], vor.vertices[simplex, 1], "r-")
                        self_voronoi.graphical_objects.append(l)

                center = points.mean(axis=0)
                for pointidx, simplex in zip(vor.ridge_points, vor.ridge_vertices):
                    simplex = np.asarray(simplex)
                    if np.any(simplex < 0):
                        i = simplex[simplex >= 0][0]  # finite end Voronoi vertex
                        t = points[pointidx[1]] - points[pointidx[0]]  # tangent
                        t /= np.linalg.norm(t)
                        n = np.array([-t[1], t[0]])  # normal
                        midpoint = points[pointidx].mean(axis=0)
                        point = vor.vertices[i]
                        far_point = point + np.sign(np.dot(midpoint - center, n)) * n * np.sqrt(
                            np.diff(xlim) ** 2 + np.diff(ylim) ** 2
                        )

                        (l,) = sup_self.ax.plot(
                            [vor.vertices[i, 0], far_point[0]],
                            [vor.vertices[i, 1], far_point[1]],
                            "r-",
                        )
                        self_voronoi.graphical_objects.append(l)

            data, x, y, _, _ = extract_data_from_mesh(self_voronoi.sup_self.ax.get_children()[0])

            x_mean = (x[1:] + x[:-1]) / 2
            y_mean = (y[1:] + y[:-1]) / 2
            x_rep = np.tile(x_mean, [len(y_mean), 1])
            y_rep = np.tile(y_mean, [len(x_mean), 1]).T

            distances_2 = []
            for p in points:
                distances_2.append((x_rep - p[0]) ** 2 + (y_rep - p[1]) ** 2)

            ind = np.argmin(distances_2, axis=0)
            sum_vec = []
            for i in range(len(points)):
                sum_vec.append(np.sum(data[ind == i]))
            print(f"points: {points}")
            print(f"population: {sum_vec}")

            plt.pause(0.0001)

        def add_marker(self_voronoi, x, y):
            self_voronoi.markers.append([x, y])
            self_voronoi.generate_voronoi()

        def mouse_click(self_voronoi, x, y):
            self_voronoi.x_click = x
            self_voronoi.y_click = y

        def move_marker(self_voronoi, x_click, y_click, x_release, y_release):
            distance = []
            for xy in self_voronoi.markers:
                distance.append((x_click - xy[0]) ** 2 + (y_click - xy[1]) ** 2)
            ind = np.argmin(distance)
            self_voronoi.markers[ind] = [x_release, y_release]
            self_voronoi.generate_voronoi()

    class MeshSelected(proto_2d_graphics):
        def __init__(self_mesh, sup_self, obj):
            self_mesh.sup_self = sup_self
            self_mesh.obj = obj
            (
                self_mesh.data,
                self_mesh.x,
                self_mesh.y,
                self_mesh.x2,
                self_mesh.y2,
            ) = extract_data_from_mesh(obj)

            self_mesh.plot = lambda x, y, data: self_mesh.sup_self.ax.pcolormesh(x, y, data)

    class PcolorSelected(proto_2d_graphics):
        def __init__(self_pcolor, sup_self, obj):
            self_pcolor.sup_self = sup_self
            self_pcolor.obj = obj
            data, x2d, y2d, x, y, x2, y2 = extract_data_from_pcolor(obj)
            self_pcolor.data = data

            self_pcolor.x = x2
            self_pcolor.y = y2
            self_pcolor.x2 = x
            self_pcolor.y2 = y
            self_pcolor.plot = lambda x, y, data: self_pcolor.sup_self.ax.pcolormesh(x, y, data)

    class LineSelected:
        def __init__(self_line, sup_self, obj, line_index_selected, xy):
            self_line.sup_self = sup_self
            self_line.obj = obj
            self_line.line_index_selected = line_index_selected
            self_line.line_width = self_line.obj.get_linewidth()

            self_line.emphasize_line_width()
            self_line.alive = True
            self_line.click_xy = xy
            if sup_self.marker_mode:
                sup_self.marker_list[-1] = Marker(sup_self.ax, self_line.click_xy)
            if hasattr(self_line.obj, "fit_text"):
                print(self_line.obj.fit_text)
                clip.OpenClipboard()
                clip.EmptyClipboard()
                clip.SetClipboardData(win32con.CF_UNICODETEXT, self_line.obj.fit_text)
                clip.CloseClipboard()

        def transpose(self_line):
            y = self_line.obj.get_xdata()
            x = self_line.obj.get_ydata()
            self_line.obj.set_xdata(x)
            self_line.obj.set_ydata(y)
            xlim = self_line.sup_self.ax.get_xlim()
            ylim = self_line.sup_self.ax.get_ylim()
            self_line.sup_self.ax.set_xlim(ylim)

            self_line.sup_self.ax.set_ylim(xlim)

        def emphasize_line_width(self_line):
            if self_line.obj.get_linestyle() == "None":
                self_line.obj.set_marker("o")
            else:
                self_line.obj.set_linewidth(self_line.line_width + 1)

        def restore_line_width(self_line):
            if self_line.obj.get_linestyle() == "None":
                self_line.obj.set_marker(".")
            else:
                self_line.obj.set_linewidth(self_line.line_width)

        def copy_to_clipboard(self_line):
            data = ""
            x_list = self_line.obj.get_xdata()
            y_list = self_line.obj.get_ydata()
            if self_line.sup_self.rectangle:
                x_list, y_list = self_line.sup_self.rectangle.filter(x_list, y_list)
                self_line.sup_self.rectangle = None
            for x, y in zip(x_list, y_list):
                data += f"{x}\t{y}\r\n"
            clip.OpenClipboard()
            clip.EmptyClipboard()
            clip.SetClipboardData(win32con.CF_UNICODETEXT, data)
            clip.CloseClipboard()

        def delete(self_line):
            if self_line.sup_self.rectangle:
                x_list = self_line.obj.get_xdata()
                y_list = self_line.obj.get_ydata()
                x_list, y_list, _ = self_line.sup_self.rectangle.filter_neg(x_list, y_list)
                self_line.obj.set_xdata(x_list)
                self_line.obj.set_ydata(y_list)
                self_line.sup_self.rectangle = None
            else:
                self_line.obj.remove()
                self_line.alive = False

        def delete_neg(self_line):
            if self_line.sup_self.rectangle:
                x_list = self_line.obj.get_xdata()
                y_list = self_line.obj.get_ydata()
                x_list, y_list, _ = self_line.sup_self.rectangle.filter(x_list, y_list)
                self_line.obj.set_xdata(x_list)
                self_line.obj.set_ydata(y_list)
                self_line.sup_self.rectangle = None

        def color(self_line, key):
            self_line.obj.set_color(key)

        def correct_order(self_line, change):

            sup_self = self_line.sup_self
            sup_self.check_zorder()
            local_order = sup_self.correct_order[self_line.line_index_selected]
            new_order = local_order + change

            if new_order > np.max(sup_self.correct_order) or new_order < np.min(sup_self.correct_order):
                return
            # test
            local_idx = self_line.line_index_selected
            to_replace_idx = np.where(np.array(sup_self.correct_order) == new_order)[0]
            if len(to_replace_idx) == 1:
                sup_self.correct_order[local_idx] = new_order
                sup_self.correct_order[to_replace_idx[0]] = local_order
            sup_self.update_zorder()

        def __del__(self_line):
            if self_line.alive:
                self_line.restore_line_width()

    def check_zorder(self):
        self.lines = self.get_lines()
        while len(self.correct_order) < len(self.lines):
            self.correct_order.append(max(self.correct_order) + 1)

    def update_zorder(self):
        for i, line in enumerate(self.lines):
            line.set_zorder(self.correct_order[i] + 2.5)

        return 0

    def get_lines(self, compare_list=["Native"]):
        self.lines = []
        lines_list_raw = self.ax.get_lines()
        for line in lines_list_raw:
            if not hasattr(line, "InteractivePlotLib_Type"):
                setattr(line, "InteractivePlotLib_Type", "Native")
            for comp in compare_list:
                if line.InteractivePlotLib_Type == comp:
                    self.lines.append(line)
                    break

        return self.lines

    def get_lines_fit_extended(self):
        return self.get_lines(compare_list=["Native", "Fit_markers", "Fit"])

    class ConvertLogLin:
        def __init__(self_loglin, type_list):  # x,y
            self_loglin.convert = []
            self_loglin.un_convert = []
            self_loglin.bound = []
            for t in type_list:
                if t == "log":
                    self_loglin.convert.append(lambda x: np.log10(np.maximum(x, 1e-16)))
                    self_loglin.bound.append(lambda x: np.maximum(x, 1e-16))
                    self_loglin.un_convert.append(lambda x: 10**x)
                elif t == "linear":
                    self_loglin.convert.append(lambda x: x)
                    self_loglin.bound.append(lambda x: x)
                    self_loglin.un_convert.append(lambda x: x)
                else:
                    print("unknown scale")

    def refresh_axis(self):
        self.ax = plt.gca()

    def refresh_figure(self):
        self.fig.canvas.draw()

    def mouse_click(self, event):
        self.user_interaction("mouse_click", event)

    def mouse_release(self, event):
        self.user_interaction("mouse_release", event)

    def keyboard_click(self, event):
        self.user_interaction("keyboard_click", event)

    def keyboard_release(self, event):
        self.user_interaction("keyboard_release", event)

    class ProtoStateMachine:
        def __init__(self_state, sup_self, done=False):
            self_state.done = done
            self_state.sup_self = sup_self

        def __del__(self_state):
            try:
                self_state.sup_self.refresh_figure()
            except BaseException:
                pass

    class StateMachineLegend(ProtoStateMachine):
        def __init__(self_state, sup_self):
            super().__init__(sup_self)
            self_state.legend = [""]
            self_state.lines = sup_self.get_lines()
            if sup_self.ax.get_legend():
                sup_self.ax.get_legend().set_visible(True)
                self_state.legend = [i.get_text() for i in sup_self.ax.get_legend().texts]
            self_state.legend_original = self_state.legend[:]
            self_state.index = 0
            self_state.text_obj = InteractivePlotLibFigure.InteractiveText(
                self_state.legend[self_state.index], self_state.update_legend
            )

        def event(self_state, type, event):
            if type == "keyboard_click":
                if event.key == "down":
                    if self_state.index < len(self_state.lines) - 1:
                        self_state.text_obj.done()
                        self_state.index += 1
                        if self_state.index >= len(self_state.legend):
                            self_state.legend.append("")
                        self_state.text_obj = InteractivePlotLibFigure.InteractiveText(
                            self_state.legend[self_state.index],
                            self_state.update_legend,
                        )

                elif event.key == "up":
                    if self_state.index > 0:
                        self_state.text_obj.done()
                        self_state.index -= 1
                        self_state.text_obj = InteractivePlotLibFigure.InteractiveText(
                            self_state.legend[self_state.index],
                            self_state.update_legend,
                        )

                else:
                    self_state.done = self_state.text_obj.react_to_key_press(event.key)

            elif type == "mouse_click":
                self_state.done = self_state.text_obj.done()
                self_state.sup_self.user_interaction("mouse_click", event)

        def update_legend(self_state, text):
            self_state.legend[self_state.index] = text
            self_state.sup_self.ax.legend(self_state.legend)

    class StateMachineLim(ProtoStateMachine):
        def __init__(self_state, sup_self, type="xlabel"):
            super().__init__(sup_self)
            self_state.type = type
            self_state.curser_location = 0
            self_state.command_stage = ""
            self_state.old_units = ""
            xlim = sup_self.ax.get_xlim()
            ylim = sup_self.ax.get_ylim()
            if type == "xstart":
                text = str(xlim[0])
                loc = [xlim[0], ylim[0]]
                va = "bottom"
                ha = "left"
            elif type == "xend":
                text = str(xlim[1])
                loc = [xlim[1], ylim[0]]
                va = "bottom"
                ha = "right"
            if type == "ystart":
                text = str(ylim[0])
                loc = [xlim[0], ylim[0]]
                va = "bottom"
                ha = "left"
            elif type == "yend":
                text = str(ylim[1])
                loc = [xlim[0], ylim[1]]
                va = "top"
                ha = "left"
            else:
                pass
            self_state.text_box = InteractivePlotLibFigure.TextObj(sup_self.ax, loc, text, ha=ha, va=va)
            self_state.text_obj = InteractivePlotLibFigure.InteractiveText(
                "", self_state.text_box.update_text, format_text=lambda x: x
            )

        def run_command(self_state, command, type):
            if self_state.command_stage == "unit_conversion":
                pass

        def event(self_state, type, event):
            if type == "keyboard_click":
                self_state.done = self_state.text_obj.react_to_key_press(event.key)

            elif type == "mouse_click":
                self_state.done = self_state.text_obj.done()
                self_state.sup_self.user_interaction("mouse_click", event)

            if self_state.done:
                xlim = self_state.sup_self.ax.get_xlim()
                ylim = self_state.sup_self.ax.get_ylim()
                try:
                    if self_state.type == "xstart":
                        self_state.sup_self.ax.set_xlim(np.sort([float(self_state.text_obj.text), xlim[1]]))
                    elif self_state.type == "xend":
                        self_state.sup_self.ax.set_xlim(np.sort([xlim[0], float(self_state.text_obj.text)]))
                    elif self_state.type == "ystart":
                        self_state.sup_self.ax.set_ylim(np.sort([float(self_state.text_obj.text), ylim[1]]))
                    elif self_state.type == "yend":
                        self_state.sup_self.ax.set_ylim(np.sort([ylim[0], float(self_state.text_obj.text)]))
                except BaseException:
                    pass

                self_state.text_box.remove()
                self_state.sup_self.refresh_figure()

    class StateMachineLabel(ProtoStateMachine):
        def __init__(self_state, sup_self, type="xlabel"):
            super().__init__(sup_self)
            self_state.type = type
            self_state.curser_location = 0
            self_state.command_stage = ""
            self_state.old_units = ""

            if type == "xlabel":
                text = sup_self.ax.get_xlabel()

                def set_label(t):
                    return sup_self.ax.set_xlabel(t)

            elif type == "ylabel":
                text = sup_self.ax.get_ylabel()

                def set_label(t):
                    return sup_self.ax.set_ylabel(t)

            elif type == "title":
                text = sup_self.ax.get_title()

                def set_label(t):
                    return sup_self.ax.set_title(t)

            text_help = "commands: \n:lin[ear]\n:log\n:u[nits]\n:[func]"
            self_state.help_box = InteractivePlotLibFigure.HelperText(self_state.sup_self.ax, text_help)

            self_state.text_obj = InteractivePlotLibFigure.InteractiveText(
                text, set_label, lambda x: x, lambda x: self_state.run_command(x, type)
            )

        def run_command(self_state, command, type):
            is_done = True
            if self_state.command_stage == "unit_conversion":
                # try:
                value_unit_split = self_state.text_obj.text.split("[")
                new_units = value_unit_split[1].split("]")[0]
                print(f"old:{self_state.old_units},new:{new_units}")
                convert_function = self_state.convert_units(self_state.old_units, new_units)
                if type == "xlabel":
                    for line in self_state.sup_self.ax.get_lines():
                        line.set_xdata(convert_function(line.get_xdata()))
                    self_state.sup_self.ax.set_xlim(convert_function(self_state.sup_self.ax.get_xlim()))
                elif type == "ylabel":
                    for line in self_state.sup_self.ax.get_lines():
                        line.set_ydata(convert_function(line.get_ydata()))
                    self_state.sup_self.ax.set_ylim(convert_function(self_state.sup_self.ax.get_ylim()))
            else:
                if command == ":lin" or command == ":linear":
                    self_state.text_obj.text = self_state.text_obj.original_text
                    if type == "xlabel":
                        self_state.sup_self.ax.set_xscale("linear")
                    elif type == "ylabel":
                        self_state.sup_self.ax.set_yscale("linear")
                    self_state.text_obj.update_text()
                elif command == ":log":
                    self_state.text_obj.text = self_state.text_obj.original_text
                    if type == "xlabel":
                        self_state.sup_self.ax.set_xscale("log")
                    elif type == "ylabel":
                        self_state.sup_self.ax.set_yscale("log")
                    self_state.text_obj.update_text()
                elif command == ":u" or command == ":units":
                    try:
                        value_unit_split = self_state.text_obj.original_text.split("[")
                        self_state.old_units = value_unit_split[1].split("]")[0]
                        self_state.text_obj.text = value_unit_split[0] + "[]"
                        self_state.text_obj.curser_location = len(self_state.text_obj.text) - 1
                        self_state.text_obj.command_mode = True
                        self_state.text_obj.update_text()
                        self_state.command_stage = "unit_conversion"
                        is_done = False
                    except BaseException:
                        self_state.text_obj.text = self_state.text_obj.original_text
                        self_state.text_obj.update_text()
                        print("can't parse units - label should be 'Amplitude [mV]'")
                else:
                    try:

                        def convert_function(x):
                            return x

                        if type == "xlabel":
                            convert_function = eval("lambda x: " + command[1:])
                            for line in self_state.sup_self.ax.get_lines():
                                line.set_xdata(convert_function(line.get_xdata()))
                            self_state.sup_self.ax.set_xlim(
                                convert_function(np.array(self_state.sup_self.ax.get_xlim()))
                            )

                        elif type == "ylabel":
                            convert_function = eval("lambda y: " + command[1:])
                            for line in self_state.sup_self.ax.get_lines():
                                line.set_ydata(convert_function(line.get_ydata()))
                            self_state.sup_self.ax.set_ylim(
                                convert_function(np.array(self_state.sup_self.ax.get_ylim()))
                            )
                        self_state.text_obj.text = self_state.text_obj.original_text
                        self_state.text_obj.update_text()
                    except BaseException:
                        print("command not parsed")

            return is_done  # mark as done

        def event(self_state, type, event):
            if type == "keyboard_click":
                self_state.done = self_state.text_obj.react_to_key_press(event.key)

            elif type == "mouse_click":
                self_state.done = self_state.text_obj.done()
                self_state.sup_self.user_interaction("mouse_click", event)

            if self_state.done:
                self_state.help_box.remove()
                self_state.sup_self.refresh_figure()

        def convert_units(self, initial, final):

            if initial.lower() == "dbm":  # to Vp

                def to_SI(x):
                    return np.sqrt(50 / 1000) * (10 ** (x / 20)) * np.sqrt(2)

            elif initial[0] == "T":

                def to_SI(x):
                    return x * 1e12

            elif initial[0] == "G":

                def to_SI(x):
                    return x * 1e9

            elif initial[0] == "M":

                def to_SI(x):
                    return x * 1e6

            elif initial[0] == "K":

                def to_SI(x):
                    return x * 1e3

            elif initial[0] == "m":

                def to_SI(x):
                    return x * 1e-3

            elif initial[0] == "u":

                def to_SI(x):
                    return x * 1e-6

            elif initial[0] == "n":

                def to_SI(x):
                    return x * 1e-9

            elif initial[0] == "p":

                def to_SI(x):
                    return x * 1e-12

            elif initial[0] == "f":

                def to_SI(x):
                    return x * 1e-15

            elif initial[0] == "a":

                def to_SI(x):
                    return x * 1e-18

            else:

                def to_SI(x):
                    return x

            if "pp" in initial.lower():

                def to_SI2(x):
                    return to_SI(x / 2)

            elif "rms" in initial.lower():

                def to_SI2(x):
                    return to_SI(x * np.sqrt(2))

            else:
                to_SI2 = to_SI

            if final.lower() == "dbm":  # assuming Vp is input

                def from_SI(x):
                    return 10 * np.log10(((x / np.sqrt(2)) ** 2) * 1000 / 50)

            elif final[0] == "T":

                def from_SI(x):
                    return x * 1e-12

            elif final[0] == "G":

                def from_SI(x):
                    return x * 1e-9

            elif final[0] == "M":

                def from_SI(x):
                    return x * 1e-6

            elif final[0] == "K":

                def from_SI(x):
                    return x * 1e-3

            elif final[0] == "m":

                def from_SI(x):
                    return x * 1e3

            elif final[0] == "u":

                def from_SI(x):
                    return x * 1e6

            elif final[0] == "n":

                def from_SI(x):
                    return x * 1e9

            elif final[0] == "p":

                def from_SI(x):
                    return x * 1e12

            elif final[0] == "f":

                def from_SI(x):
                    return x * 1e15

            elif final[0] == "a":

                def from_SI(x):
                    return x * 1e18

            else:

                def from_SI(x):
                    return x

            if "pp" in final.lower():

                def from_SI2(x):
                    return from_SI(x * 2)

            elif "rms" in final.lower():

                def from_SI2(x):
                    return from_SI(x / np.sqrt(2))

            else:
                from_SI2 = from_SI

            return lambda x: from_SI2(to_SI2(np.array(x)))

    class StateMachineCommand(ProtoStateMachine):
        def __init__(self_state, sup_self):
            super().__init__(sup_self)
            self_state.text_box = InteractivePlotLibFigure.CommandText(sup_self.ax)
            self_state.text_obj = InteractivePlotLibFigure.InteractiveText(
                "",
                self_state.text_box.update_text,
                format_text=lambda x: f"command: {x}",
            )

        def event(self_state, type, event):
            if type == "keyboard_click":
                self_state.done = self_state.text_obj.react_to_key_press(event.key)
            if type == "mouse_click":
                self_state.done = self_state.text_obj.done()
            if self_state.done:
                if self_state.text_obj.text == "colorbar":
                    if self_state.sup_self.plot_type == "mesh" or self_state.sup_self.plot_type == "pcolor":
                        plt.colorbar()

                if self_state.text_obj.text == "log":
                    if self_state.sup_self.plot_type == "mesh" or self_state.sup_self.plot_type == "pcolor":
                        self_state.sup_self.line_selected.obj.set_norm(Colors.LogNorm())
                    elif self_state.sup_self.plot_type == "plot":
                        self_state.sup_self.ax.set_yscale("log")

                if self_state.text_obj.text == "lin" or self_state.text_obj.text == "linear":
                    if self_state.sup_self.plot_type == "mesh" or self_state.sup_self.plot_type == "pcolor":
                        self_state.sup_self.line_selected.obj.set_norm(Colors.Normalize())
                    elif self_state.sup_self.plot_type == "plot":
                        self_state.sup_self.ax.set_yscale("linear")

    class StateMachineColor(ProtoStateMachine):
        def __init__(self_state, sup_self):
            super().__init__(sup_self)
            self_state.text_box = InteractivePlotLibFigure.CommandText(sup_self.ax)
            self_state.text_obj = InteractivePlotLibFigure.InteractiveText(
                "", self_state.text_box.update_text, format_text=lambda x: f"color: {x}"
            )

        def event(self_state, type, event):
            if type == "keyboard_click" or type == "mouse_click":
                self_state.done = self_state.text_obj.done()
                if type == "keyboard_click":
                    self_state.sup_self.line_selected.color(event.key)
                else:
                    self_state.sup_self.user_interaction("mouse_click", event)

    class Rectangle:
        def __init__(self, ax, x, y):
            self.x = np.sort(x)
            self.y = np.sort(y)
            rect = patches.Rectangle(
                (self.x[0], self.y[0]),
                self.x[1] - self.x[0],
                self.y[1] - self.y[0],
                edgecolor="r",
                facecolor="none",
            )
            print(f"x:[{self.x[0]} , {self.x[1]}] \n y:[{self.y[0]} , {self.y[0]}]")
            self.obj = ax.add_patch(rect)

        def filter(self, x, y):
            ind_bool = (x >= self.x[0]) * (x <= self.x[1]) * (y >= self.y[0]) * (y <= self.y[1])
            return np.array(x)[ind_bool], np.array(y)[ind_bool], ind_bool

        def filter_neg(self, x, y):
            ind_bool = (x <= self.x[0]) | (x >= self.x[1]) | (y <= self.y[0]) | (y >= self.y[1])
            return np.array(x)[ind_bool], np.array(y)[ind_bool], ind_bool

        def __del__(self):
            self.obj.remove()

    class StateMachineVoronoi(ProtoStateMachine):
        def __init__(self_state, sup_self):
            super().__init__(sup_self)
            text = "click to add node\n click&drag to move"
            self_state.help_box = InteractivePlotLibFigure.HelperText(self_state.sup_self.ax, text)
            self_state.x = [None, None]
            self_state.y = [None, None]

        def event(self_state, type, event):

            if type == "keyboard_click":
                self_state.done = True  # any key will stop voronoi

            if type == "mouse_click":

                self_state.x[0] = event.xdata
                self_state.y[0] = event.ydata

            if type == "mouse_release":

                self_state.x[1] = event.xdata
                self_state.y[1] = event.ydata
                if self_state.x[1] == self_state.x[0] and self_state.y[1] == self_state.y[0]:
                    self_state.sup_self.voronoi_obj.add_marker(self_state.x[0], self_state.y[0])
                else:
                    self_state.sup_self.voronoi_obj.move_marker(
                        self_state.x[0],
                        self_state.y[0],
                        self_state.x[1],
                        self_state.y[1],
                    )

            if self_state.done:
                self_state.help_box.remove()

    class StateMachineRect(ProtoStateMachine):
        def __init__(self_state, sup_self):
            super().__init__(sup_self)
            text = "Select rectangle\nclick & drag"
            self_state.help_box = InteractivePlotLibFigure.HelperText(self_state.sup_self.ax, text)
            self_state.x = [None, None]
            self_state.y = [None, None]

        def event(self_state, type, event):

            if type == "mouse_click":
                self_state.x[0] = event.xdata
                self_state.y[0] = event.ydata

            if type == "mouse_release":
                self_state.x[1] = event.xdata
                self_state.y[1] = event.ydata
                if self_state.x[1] == self_state.x[0] and self_state.y[1] == self_state.y[0]:

                    self_state.x = self_state.sup_self.ax.get_xlim()
                    if self_state.sup_self.plot_type == "pcolor":
                        _, _, _, _, _, _, y_ext = extract_data_from_pcolor(
                            self_state.sup_self.line_selected.obj  # plt.gca().get_children()[0]
                        )
                        mx = y_ext[y_ext <= self_state.y[0]].max()
                        mm = y_ext[y_ext >= self_state.y[0]].min()
                        self_state.y = [
                            mm - (mx - mm) / 3,
                            mx + (mx - mm) / 3,
                        ]

                self_state.sup_self.rectangle = InteractivePlotLibFigure.Rectangle(
                    self_state.sup_self.ax, self_state.x, self_state.y
                )
                self_state.done = True

            if self_state.done:
                self_state.help_box.remove()

    class StateMachineFit(ProtoStateMachine):
        def __init__(self_state, sup_self):
            super().__init__(sup_self)
            text = "x*a[0]+a[1],[1,1]\n"
            text += "# - polynom\n"
            text += "g - gaussian\n"
            text += "l - lorenzian\n"
            text += "e - exponent\n"
            text += "r - erf\n"
            text += "s - sine\n"
            text += "S - decay sine\n"

            self_state.text_box = InteractivePlotLibFigure.CommandText(sup_self.ax)
            self_state.text_obj = InteractivePlotLibFigure.InteractiveText(
                "",
                self_state.text_box.update_text,
                format_text=lambda x: f"Fit type: {x}",
            )
            self_state.help_box = InteractivePlotLibFigure.HelperText(self_state.sup_self.ax, text)

        def event(self_state, type, event):
            if type == "keyboard_click":
                self_state.done = self_state.text_obj.react_to_key_press(event.key)

            elif type == "mouse_click":
                self_state.done = self_state.text_obj.done()
                self_state.sup_self.user_interaction("mouse_click", event)

            if self_state.done:
                self_state.help_box.remove()
                print(self_state.text_obj.text)
                if self_state.sup_self.plot_type == "plot":
                    InteractivePlotLibFigure.predefined_fit(self_state.sup_self, self_state.text_obj.text)
                elif self_state.sup_self.plot_type == "pcolor" or self_state.sup_self.plot_type == "mesh":
                    InteractivePlotLibFigure.predefined_fit_pcolor(
                        self_state.sup_self,
                        self_state.text_obj.text,
                        self_state.sup_self.plot_type,
                    )
                self_state.sup_self.refresh_figure()

    class InteractiveText:
        # :
        def __init__(
            self_text,
            initial_text,
            update_text_fun,
            format_text=lambda x: x,
            command_func=lambda x: [],
        ):
            self_text.text = initial_text
            self_text.original_text = initial_text
            self_text.format_text = format_text
            self_text.update_text_fun = update_text_fun
            self_text.is_done = False
            self_text.curser_location = len(self_text.text)
            self_text.command_func = command_func
            self_text.update_text()
            self_text.first_key_stroke = True
            self_text.command_mode = False

        def update_text(self_text):
            buf = self_text.text
            if not self_text.is_done:
                buf = buf[: self_text.curser_location] + "|" + buf[self_text.curser_location :]
            self_text.update_text_fun(self_text.format_text(buf))
            self_text.first_key_stroke = False

        def done(self_text):
            self_text.is_done = True
            self_text.update_text()
            return self_text.is_done

        def react_to_key_press(self_text, key):
            # print(self_text.first_key_stroke)
            if key == "enter":
                if (len(self_text.text) > 0 and self_text.text[0] == ":") or self_text.command_mode:
                    is_done = self_text.command_func(self_text.text)
                    if is_done:
                        return self_text.done()
                    else:
                        return is_done  # =False

                return self_text.done()

            elif key == "escape":
                self_text.text = self_text.original_text
                return self_text.done()

            elif key == "backspace":
                if (len(self_text.text) > 0) and (self_text.curser_location > 0):
                    self_text.text = (
                        self_text.text[: self_text.curser_location - 1] + self_text.text[self_text.curser_location :]
                    )
                    self_text.curser_location -= 1

            elif key == ":" and self_text.first_key_stroke:
                self_text.curser_location = 1
                self_text.text = ":"

            elif key == "delete":
                self_text.curser_location = 0
                self_text.text = ""

            elif key == "left":
                if self_text.curser_location > 0:
                    self_text.curser_location -= 1

            elif key == "right":
                if self_text.curser_location < len(self_text.text):
                    self_text.curser_location += 1

            elif len(key) == 1:
                self_text.text = (
                    self_text.text[: self_text.curser_location] + key + self_text.text[self_text.curser_location :]
                )
                self_text.curser_location += 1

            else:
                return self_text.is_done

            self_text.update_text()

            return self_text.is_done

    class TextObj:
        def __init__(self_textobj, ax, loc, text="", ha="left", va="top"):
            print(loc)
            self_textobj.text_obj = ax.text(
                loc[0],
                loc[1],
                text,
                ha=ha,
                va=va,
            )

        def update_text(self_textobj, text):
            self_textobj.text_obj.set_text(text)

        def remove(self_textobj):
            self_textobj.text_obj.remove()

        def __del__(self_textobj):
            try:
                self_textobj.text_obj.remove()
            except BaseException:
                pass

    class CommandText(TextObj):
        def __init__(self, ax, text=""):
            super().__init__(ax, (min(ax.get_xlim()), max(ax.get_ylim())), text)

    class HelperText(TextObj):
        def __init__(self, ax, text=""):
            super().__init__(ax, (max(ax.get_xlim()), max(ax.get_ylim())), text, ha="right")


class Marker:
    def __init__(self_marker, ax, loc, text="", color=".g"):
        if not text:
            x_text = f"{loc[0]:0.3f}" if ((np.abs(loc[0]) > 1e-2) and (np.abs(loc[0]) < 1e2)) else f"{loc[0]:0.2e}"
            y_text = f"{loc[1]:0.3f}" if ((np.abs(loc[1]) > 1e-2) and (np.abs(loc[1]) < 1e2)) else f"{loc[1]:0.2e}"
            text = f"[{x_text},{y_text}]"
        self_marker.text = InteractivePlotLibFigure.TextObj(ax, loc, text, "left", "bottom")
        self_marker.text.text_obj.set_bbox(dict(facecolor="white", alpha=0.5, edgecolor="none"))
        (self_marker.point,) = ax.plot(loc[0], loc[1], color)
        setattr(self_marker.point, "InteractivePlotLib_Type", "Marker")
        plt.pause(0.001)

    def remove(self_marker):
        try:
            self_marker.point.remove()
            self_marker.text.remove()
        except BaseException:
            pass

    def __del__(self_marker):
        self_marker.remove()


class Document:
    def __init__(self, Data_path):
        self.Data_path = Data_path

    def doc(self, figs, unfiltered_variables=[], ignore=[]):
        print(figs)
        if not unfiltered_variables:
            unfiltered_variables = {}

            unfiltered_variables_module = _main_module

            att_list = dir(unfiltered_variables_module)

            for att in att_list:

                if att[0] == "_":
                    continue
                unfiltered_variables[att] = getattr(unfiltered_variables_module, att)

        self.ignore = ignore

        scan_number = self.save(unfiltered_variables)
        self.scan_number = scan_number
        print(type(figs))
        if type(figs) is list or type(figs) is np.ndarray:
            for fig in figs:
                self.save_fig(fig_number=fig, scan_number=scan_number)

        else:
            self.save_fig(fig_number=figs, scan_number=scan_number)

    def white_list(self, x, var):
        if x[0] == "_":
            return []
        if type(var) not in [
            np.ndarray,
            list,
            float,
            int,
            str,
            dict,
            complex,
            tuple,
        ]:
            return []
        if x in ["In", "Out", "exit", "get_ipython", "quit"]:
            return []
        if x in self.ignore:
            return []

        return x

    def get_variables(self, unfiltered_variables):
        out = []

        for x in sorted(unfiltered_variables.keys()):
            to_append = self.white_list(x, unfiltered_variables[x])
            if to_append:
                out.append(to_append)
        return out

    def get_last_scan_number(self):
        file_list = os.listdir(self.Data_path)

        if not file_list:
            scan_number = -1
        else:
            scans_list = [-1]
            for x in file_list:
                if "scan" in x:
                    scans_list.append(int(x.split("_")[0].split("scan")[1]))
            scan_number = max(scans_list)
        return scan_number

    def save(self, unfiltered_variables):

        Data_path = self.Data_path
        try:
            os.remove("./temp_code.txt")
        except BaseException:
            pass

        ipython.magic('history -n -f "./temp_code.txt"')
        f = open("./temp_code.txt", "r")
        code = f.read()
        f.close()

        scan_number = self.get_last_scan_number() + 1
        now = datetime.datetime.now()
        current_time = now.strftime("%Y_%m_%d_%H_%M_%S")
        info = {}
        variables = self.get_variables(unfiltered_variables)
        for variable in variables:
            info[variable] = unfiltered_variables[variable]
            if type(info[variable]) is list:
                for i in range(len(info[variable])):
                    if str(type(info[variable][i])).split("'")[1].split(".")[0] == "matplotlib":
                        info[variable][i] = []
        info["code"] = code
        dill.dump(
            info,
            open((Data_path + "/scan{}_{}.dat").format(scan_number, current_time), "wb"),
        )
        print(variables)
        return scan_number

    def load(self, n, load_figures=True):
        Data_path = self.Data_path

        if load_figures:

            try:
                if isinstance(n, int):
                    figures_path = glob.glob(Data_path + "\\scan" + str(n) + "_*.pkl")
                else:
                    figures_path = [n]

                for figure_path in figures_path:
                    dill.load(open(figure_path, "rb"))

            except BaseException:
                print("problem loading figures")

        if isinstance(n, int):
            data_path = glob.glob(Data_path + "\\scan" + str(n) + "_*.dat")
        else:
            data_path = glob.glob(n.split("fig")[0] + "*.dat")

            print(data_path)

        if len(data_path) > 0:
            loaded_data = dill.load(open(data_path[0], "rb"))
        else:
            print("no such scan!")
            return 0

        print(loaded_data.keys())

        return loaded_data

    def save_fig(self, fig_number, scan_number=0):
        path = self.Data_path

        name = (
            "scan"
            + str(scan_number)
            + "_fig"
            + str(fig_number)
            + "_"
            + datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        )
        path = os.path.join(path, name)

        h = plt.figure(fig_number)
        for t in h.axes[0].texts:
            if len(t.get_text()) > 5 and t.get_text()[:5] == "Scan ":

                t.remove()

        plt.text(
            plt.gca().get_xlim()[0],
            plt.gca().get_ylim()[1],
            "Scan " + str(self.scan_number),
            fontsize=12,
            va="top",
        )
        plt.pause(0.01)
        h.savefig(path + ".png")

        dill.dump(h, open(path + ".pkl", "wb"))

        self.generate_csv_data(h, path)

        image = ImageGrab.grab()
        output = BytesIO()
        mngr = plt.get_current_fig_manager()
        rect = mngr.window.geometry().getRect()
        image = image.crop((rect[0], rect[1], rect[0] + rect[2], rect[1] + rect[3]))

        image.convert("RGB").save(output, "BMP")
        data = output.getvalue()[14:]
        output.close()
        clip.OpenClipboard()
        clip.EmptyClipboard()
        clip.SetClipboardData(win32con.CF_DIB, data)
        clip.CloseClipboard()

        return path

    def generate_csv_data(self, fig, path):
        data = []
        ax = fig.get_axes()[0]
        for line in ax.get_lines():
            data.append({"x": line.get_xdata(), "y": line.get_ydata()})
        leg = ax.get_legend()
        if leg is not None:
            leg_names = []
            for t in ax.get_legend().get_texts():
                leg_names.append(t.get_text())
        else:
            leg_names = ["line " + str(i + 1) for i in range(len(data))]

        title = ax.get_title()
        if title == "":
            title = "untitled"

        xlabel = ax.get_xlabel()
        if xlabel == "":
            xlabel = "x"

        ylabel = ax.get_ylabel()
        if ylabel == "":
            ylabel = "y"

        all_x_data = [j for i in data for j in i["x"]]
        all_y_data = [j for i in data for j in i["y"]]
        all_leg_names = []

        for i in range(len(data)):
            if i < len(leg_names):
                all_leg_names += [leg_names[i]] * len(data[i]["x"])
            else:
                all_leg_names += [f"line {i}"] * len(data[i]["x"])
        df = pd.DataFrame(
            {
                "title": [title] * len(all_x_data),
                "legend": all_leg_names,
                xlabel: all_x_data,
                ylabel: all_y_data,
            }
        )
        df.to_csv(path + ".csv")


class Fit:
    def __init__(self, fit_type=1, ax=[], options={}):
        self.res = []
        if not ax:
            ax = plt.gca()
        self.ax = ax
        lines = self.ax.get_lines()
        if type(fit_type) is int:
            self.res = {
                "fit_type": [],
                "fit_func": [],
                "x": [],
                "y": [],
                "yf": [],
                "p": [],
                "y0": [],
                "x0": [],
                "dp": [],
                "dx0": [],
                "dy0": [],
                "ddp": [],
                "ddx0": [],
                "ddy0": [],
            }
        elif fit_type.__code__.co_argcount == 2:
            self.res = {
                "fit_type": [],
                "fit_func": [],
                "x": [],
                "y": [],
                "yf": [],
                "a": [],
                "func": [],
            }
        else:
            self.res = {
                "fit_type": [],
                "fit_func": [],
                "x": [],
                "y": [],
                "yf": [],
                "x_shift": [],
                "y_shift": [],
                "x_scale": [],
                "y_scale": [],
                "y_mid": [],
            }
        order_list = []
        for line in lines:
            order_list.append(line.zorder)
        idx = range(len(lines))
        for i in idx:
            if hasattr(lines[i], "InteractivePlotLib_Type") and (
                lines[i].InteractivePlotLib_Type == "Fit"
                or lines[i].InteractivePlotLib_Type == "Fit_markers"
                or lines[i].InteractivePlotLib_Type == "Marker"
            ):
                continue
            single_fit = SingleFit(lines[i], fit_type, ax, options)
            if not single_fit.ok:
                continue
            self.res["fit_type"].append(single_fit.fit_type)
            self.res["fit_func"].append(single_fit.fit_func)
            self.res["x"].append(single_fit.x)
            self.res["y"].append(single_fit.y)
            self.res["yf"].append(single_fit.yf)
            if type(fit_type) is int:
                self.res["p"].append(single_fit.p)
                self.res["y0"].append(single_fit.y0)
                self.res["x0"].append(single_fit.x0)
                self.res["dp"].append(single_fit.dp)
                self.res["dx0"].append(single_fit.dx0)
                self.res["dy0"].append(single_fit.dy0)
                self.res["ddp"].append(single_fit.ddp)
                self.res["ddx0"].append(single_fit.ddx0)
                self.res["ddy0"].append(single_fit.ddy0)
            elif fit_type.__code__.co_argcount == 2:
                self.res["a"].append(single_fit.popt)
                self.res["func"].append(options["func"])
            else:
                self.res["x_shift"].append(single_fit.x_shift)
                self.res["y_shift"].append(single_fit.y_shift)
                self.res["x_scale"].append(single_fit.x_scale)
                self.res["y_scale"].append(single_fit.y_scale)
                self.res["y_mid"].append(single_fit.y_mid)

            setattr(single_fit.fit_line_obj, "print", single_fit.text)
            setattr(single_fit.fit_line_obj, "fit_func", single_fit.fit_func)
            setattr(single_fit.fit_line_obj, "x_original", lines[i].get_xdata())
            setattr(single_fit.fit_line_obj, "y_original", lines[i].get_ydata())

        setattr(self.ax, "fit", self.res)
        if hasattr(single_fit, "fit_line_markers_obj"):
            setattr(self.ax, "fit_markers", single_fit.fit_line_markers_obj)
        setattr(_main_module, "fit", self.res)

    def __repr__(self):
        return str(self.res)


class SingleFit:
    def __init__(self, line, fit_type=1, ax=[], options={}):
        if not ax:
            ax = plt.gca()
        self.fit_type = fit_type
        self.ax = ax
        self.fit_func = lambda x: np.zeros_like(x)
        self.x = ([],)
        self.y = ([],)
        self.yf = ([],)
        self.x_shift = (0,)
        self.y_shift = (0,)
        self.x_scale = (1,)
        self.y_scale = (1,)
        self.y_mid = (0,)

        if callable(fit_type):

            if fit_type.__code__.co_argcount == 1:
                self.ok = self.fit_scaled_function(line, options)

            if fit_type.__code__.co_argcount == 2:
                x_data, y_data = self.get_xdata_ydata_from_axes(line)
                x_scale, y_scale = 1, 1

                if "y_scale" in options:
                    y_scale = options["y_scale"]
                if "x_scale" in options:
                    x_scale = options["y_scale"]

                popt = curve_fit3(
                    lambda x, a: fit_type(x, a),
                    x_data / x_scale,
                    y_data / y_scale,
                    options["initial_guess"],
                )
                print(f"func = {options['func']}, a = {popt}")
                self.fit_func = lambda x: fit_type(x / x_scale, popt) * y_scale
                (l,) = plt.plot(x_data, self.fit_func(x_data), "m", linewidth=2)
                self.fit_line_obj = l
                setattr(self.fit_line_obj, "InteractivePlotLib_Type", "Fit")
                self.ok = True
                self.popt = popt
                if "text" in options:
                    self.text = options["text"](popt)
                else:
                    self.text = f"a = {popt}"
                print(self.text)
                self.x = x_data
                self.y = y_data
                self.yf = self.fit_func(x_data)
            else:
                pass

        elif type(fit_type) is int:
            self.ok = self.fit_polygon(line, options)

        else:
            pass

    def get_xdata_ydata_from_axes(self, line):

        ax = self.ax

        xlim = ax.get_xlim()
        ylim = ax.get_ylim()

        x_data = np.array(line.get_xdata())
        y_data = np.array(line.get_ydata())

        idx = (x_data >= xlim[0]) & (x_data <= xlim[1]) & (y_data >= ylim[0]) & (y_data <= ylim[1])
        x_data = x_data[idx]
        y_data = y_data[idx]
        return x_data, y_data

    def plot_points_of_interest(self, points_of_interest_x, points_of_interest_y):

        xlim = plt.gca().get_xlim()
        ylim = plt.gca().get_ylim()
        self.fit_line_markers_obj = []
        for x, y in zip(points_of_interest_x, points_of_interest_y):
            self.fit_line_markers_obj.append(Marker(plt.gca(), [x, y], color=".r"))
        for marker in self.fit_line_markers_obj:
            setattr(marker.point, "InteractivePlotLib_Type", "Fit_markers")
            setattr(
                marker.point,
                "InteractivePlotLib_Fit_Parent",
                self.fit_line_obj,
            )
        plt.gca().set_xlim(xlim)
        plt.gca().set_ylim(ylim)

    def fit_polygon(self, line, options):
        x_data, y_data = self.get_xdata_ydata_from_axes(line)
        if len(x_data) == 0:
            return False
        poly = np.polyfit(x_data, y_data, self.fit_type)
        self.p = poly
        self.fit_func = lambda x: np.polyval(poly, x)

        (l,) = plt.plot(x_data, self.fit_func(x_data), "m", linewidth=2)
        self.fit_line_obj = l
        setattr(self.fit_line_obj, "InteractivePlotLib_Type", "Fit")
        self.x = x_data
        self.y = y_data
        self.yf = self.fit_func(x_data)
        self.y0 = self.fit_func(0)
        self.x0 = np.array(np.real([i for i in np.roots(poly) if np.isreal(i)]))
        self.dp = np.polyder(self.p)
        self.dx0 = np.array(np.real([i for i in np.roots(self.dp) if np.isreal(i)]))
        self.dy0 = [self.fit_func(i) for i in self.dx0]
        self.ddp = np.polyder(self.dp)
        self.ddx0 = np.array(np.real([i for i in np.roots(self.ddp) if np.isreal(i)]))
        self.ddy0 = [self.fit_func(i) for i in self.ddx0]
        points_of_interest_x = np.hstack([0, self.x0, self.dx0, self.ddx0])
        points_of_interest_y = np.hstack([self.y0, [0] * len(self.x0), self.dy0, self.ddy0])

        self.plot_points_of_interest(points_of_interest_x, points_of_interest_y)

        funcString = f"poly {self.fit_type}"
        text = f"func: {funcString}\n"

        text += "x0:"
        for i in self.x0:
            text += f"{i:2.2e},"
        text = text[:-1]
        text += f"\ny0:{self.y0}\n"
        if self.fit_type > 1:
            text += "d(x0,y0):"
            for dx0, dy0 in zip(self.dx0, self.dy0):
                text += f"({dx0:2.2e},{dy0:2.2e}),"
            text = text[:-1]
        text += f" poly:{self.p}\n"

        if self.fit_type == 0:
            self.mean = poly[0]
            text += f"mean: {self.mean:2.2e}\n"

        if self.fit_type == 1:
            self.slope = poly[0]
            text += f"slope: {self.slope:2.2e}\n"

        self.text = text
        setattr(self.fit_line_obj, "fit_text", text)
        setattr(self.fit_line_obj, "fit_func", self.fit_func)
        print(text)
        return True

    def fit_scaled_function(self, line, options):
        func = self.fit_type

        ax = self.ax

        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        x_data, y_data = self.get_xdata_ydata_from_axes(line)
        if len(x_data) == 0:
            return False

        x_dummy = np.linspace(-np.pi / 2, np.pi / 2, 11)
        y_dummy = func(x_dummy)
        y_dummy_max = max(y_dummy)
        y_dummy_min = min(y_dummy)

        def scale_and_fit_function(f, x, shift_x, scale_x, shift_y, scale_y):
            return (f((x - shift_x) / scale_x) - shift_y) / scale_y

        shift_data_x = np.mean(xlim)
        shift_data_y = np.mean(ylim)

        scale_data_x = xlim[1] - xlim[0]
        scale_data_y = ylim[1] - ylim[0]

        scaled_x = (x_data - shift_data_x) / scale_data_x
        scaled_y = (y_data - shift_data_y) / scale_data_y

        mean_function = (y_dummy_max + y_dummy_min) / 2
        delta_function = y_dummy_max - y_dummy_min
        if "x_scale" in options:
            scaling_x_constant = options["x_scale"]
        else:
            scaling_x_constant = 0.1

        def R2(y, y_fit):
            ss_res = np.sum((y - y_fit) ** 2)
            ss_tot = np.sum((y - np.mean(y)) ** 2)
            r2 = 1 - (ss_res / ss_tot)
            return r2

        def even_or_odd(func):
            y1 = func(1)
            y2 = func(-1)
            if y1 == y2:
                return "even"
            elif y1 == -y2:
                return "odd"
            else:
                return "other"

        function_symmetry_type = even_or_odd(func)
        if function_symmetry_type == "even":
            initial_guess = [
                [0, scaling_x_constant, mean_function, delta_function],
                [0, scaling_x_constant, mean_function, -delta_function],
            ]
        elif function_symmetry_type == "odd":
            initial_guess = [
                [0, scaling_x_constant, mean_function, delta_function],
                [0, -scaling_x_constant, mean_function, delta_function],
            ]
        else:
            initial_guess = [
                [0, scaling_x_constant, mean_function, delta_function],
                [0, -scaling_x_constant, mean_function, delta_function],
                [0, scaling_x_constant, mean_function, -delta_function],
                [0, -scaling_x_constant, mean_function, -delta_function],
            ]
        candidate_list = []
        for guess in initial_guess:
            popt = curve_fit2(
                lambda x, *args: scale_and_fit_function(func, x, *args),
                scaled_x,
                scaled_y,
                guess,
            )

            candidate_list.append(
                {
                    "popt": popt,
                    "R2": R2(scaled_y, scale_and_fit_function(func, scaled_x, *popt)),
                }
            )

        idx = np.argmin([1 - i["R2"] for i in candidate_list])
        popt = candidate_list[idx]["popt"]

        output = [
            popt[0] * scale_data_x + shift_data_x,
            popt[1] * scale_data_x,
            -(scale_data_y / (popt[3]) * popt[2] - shift_data_y),
            scale_data_y / popt[3],
        ]

        self.fit_func = lambda x: func((x - output[0]) / output[1]) * output[3] + output[2]
        (l,) = plt.plot(x_data, self.fit_func(x_data), "m", linewidth=2)
        self.fit_line_obj = l
        setattr(self.fit_line_obj, "InteractivePlotLib_Type", "Fit")
        self.x = x_data
        self.y = y_data
        self.yf = self.fit_func(x_data)
        self.x_shift = output[0]
        self.y_shift = output[2]
        self.x_scale = output[1]
        self.y_scale = output[3]
        self.y_mid = self.fit_func(0)
        points_of_interest_x = np.hstack([self.x_shift, 0])
        points_of_interest_y = np.hstack([self.fit_func(self.x_shift), self.fit_func(0)])
        self.plot_points_of_interest(points_of_interest_x, points_of_interest_y)

        try:
            funcString = str(inspect.getsourcelines(func)[0])
            funcString = funcString.strip("['\\n']").split(" = ")[1]
            funcString = funcString.split(":")[1][1:]
        except BaseException:
            funcString = "?"

        text = f"func: {funcString}\n"
        text += f"→x:{self.x_shift:2.2e},↑y:{self.y_shift:2.2e}\n"
        text += f"↔x:{self.x_scale:2.2e},↕y:{self.y_scale:2.2e}\n"
        self.text = text
        setattr(self.fit_line_obj, "fit_text", text)
        print(text)
        return True


def curve_fit2(f, x, y, a0):
    def opt(x, y, a):
        return np.sum(np.abs(f(x, *a) - y) ** 2)

    out = optimize.minimize(lambda a: opt(x, y, a), a0)
    return out["x"]


def curve_fit3(f, x, y, a0):
    def opt(x, y, a):
        return np.sum(np.abs(f(x, a) - y) ** 2)

    out = optimize.minimize(lambda a: opt(x, y, a), a0)
    return out["x"]


def line_in_lim(point0, slope, xlim, ylim):
    def line(x):
        return (x - point0[0]) * slope + point0[1]

    def inv_line(y):
        return (y - point0[1]) / slope + point0[0]

    xy = []
    for x in [min(xlim), max(xlim)]:
        if line(x) > max(ylim):
            xy.append([inv_line(max(ylim)), max(ylim)])
        elif line(x) < min(ylim):
            xy.append([inv_line(min(ylim)), min(ylim)])
        else:
            xy.append([x, line(min(xlim))])
    return xy



# VideoMode script enabling the update of parameters in a QUA program through user input while the program is running.
# Authors: Arthur Strauss, Théo Laudat
# Date: 19/12/2023

import threading
from qm.qua import *
from qm import QuantumMachine, QmJob, Program
import time
import signal
from typing import Optional, List, Dict, Union, Tuple
import numpy as np


def set_type(qua_type: Union[str, type]):
    """
    Set the type of the QUA variable to be declared.
    Args: qua_type: Type of the QUA variable to be declared (int, fixed, bool).
    """

    if qua_type == "fixed" or qua_type == fixed:
        return fixed
    elif qua_type == "bool" or qua_type == bool:
        return bool
    elif qua_type == "int" or qua_type == int:
        return int
    else:
        raise ValueError("Invalid QUA type. Please use 'fixed', 'int' or 'bool'.")


def infer_type(value: Union[int, float, List, np.ndarray] = None):
    """
    Infer automatically the type of the QUA variable to be declared from the type of the initial parameter value.
    """
    if isinstance(value, float):
        if value.is_integer() and value > 8:
            return int
        else:
            return fixed

    elif isinstance(value, bool):
        return bool

    elif isinstance(value, int):
        return int

    elif isinstance(value, (List, np.ndarray)):
        if isinstance(value, np.ndarray):
            assert value.ndim == 1, "Invalid parameter type, array must be 1D."
            value = value.tolist()
        assert all(
            isinstance(x, type(value[0])) for x in value
        ), "Invalid parameter type, all elements must be of same type."
        if isinstance(value[0], bool):
            return bool
        elif isinstance(value[0], int):
            return int
        elif isinstance(value[0], float):
            return fixed
    else:
        raise ValueError("Invalid parameter type. Please use float, int or bool or list.")


class ParameterValue:
    def __init__(
        self,
        name: str,
        value: Union[int, float, List, np.ndarray],
        index: int,
        qua_type: Optional[Union[str, type]] = None,
    ):
        """ """
        self.name = name
        self.value = value
        self.index = index
        self.var = None
        self.type = set_type(qua_type) if qua_type is not None else infer_type(value)
        self.length = 0 if not isinstance(value, (List, np.ndarray)) else len(value)

    def __repr__(self):
        return f"{self.name}: ({self.value}, {self.type}) \n"


class ParameterTable:
    """
    Class enabling the mapping of parameters to be updated to their corresponding "to-be-declared" QUA variables. The
    type of the QUA variable to be adjusted is automatically inferred from the type of the initial_parameter_value.
    Each parameter in the dictionary should be given a name that the user can then easily access through the table
    with table[parameter_name]. Calling this will return the QUA variable built within the QUA program corresponding
    to the parameter name and its associated Python initial value. Args: parameters_dict: Dictionary of the form {
    "parameter_name": initial_parameter_value }. the QUA program.
    """

    def __init__(
        self,
        parameters_dict: Dict[
            str,
            Union[
                Tuple[Union[float, int, bool, List, np.ndarray], Optional[str]],
                Union[float, int, bool, List, np.ndarray],
            ],
        ],
    ):
        """
        Class enabling the mapping of parameters to be updated to their corresponding "to-be-declared" QUA variables.
        The type of the QUA variable to be adjusted can be specified or either be automatically inferred from the
        type of the initial_parameter_value. Each parameter in the dictionary should be given a name that the user
        can then easily access through the table with table[parameter_name]. Calling this will return the QUA
        variable built within the QUA program corresponding to the parameter name and its associated Python initial
        value. Args: parameters_dict: Dictionary should be of the form { "parameter_name": (initial_value,
        qua_type) } where qua_type is the type of the QUA variable to be declared (int, fixed, bool).


        """
        self.parameters_dict = parameters_dict
        self.table = {}
        for index, (parameter_name, parameter_value) in enumerate(self.parameters_dict.items()):
            if isinstance(parameter_value, Tuple):
                if len(parameter_value) != 2:
                    raise ValueError(
                        "Invalid format for parameter value. Please use (initial_value, qua_type) or initial_value."
                    )
                self.table[parameter_name] = ParameterValue(
                    parameter_name, parameter_value[0], index, parameter_value[1]
                )
            else:
                self.table[parameter_name] = ParameterValue(parameter_name, parameter_value, index)

    def declare_variables(self):
        """
        QUA Macro to create the QUA variables associated with the parameter table.
        """
        for parameter_name, parameter in self.table.items():
            parameter.var = declare(t=parameter.type, value=parameter.value)
        pause()
        if len(self.variables) == 1:
            return self.variables[0]
        else:
            return self.variables

    def load_parameters(self, pause_program=False):
        """QUA Macro to be called within QUA program to retrieve updated values for the parameters through IO 1 and IO2.
        Args:
            pause_program: Boolean indicating whether the program should be paused while waiting for user input.
        """

        if pause_program:
            pause()

        param_index_var = declare(int)
        assign(param_index_var, IO1)

        with switch_(param_index_var):
            for parameter in self.table.values():
                with case_(parameter.index):
                    if parameter.length == 0:
                        assign(parameter.var, IO2)
                    else:
                        looping_var = declare(int)
                        with for_(
                            looping_var,
                            0,
                            looping_var < parameter.var.length(),
                            looping_var + 1,
                        ):
                            pause()
                            assign(parameter.var[looping_var], IO2)

            with default_():
                pass

    def print_parameters(self):
        text = ""
        for parameter_name, parameter in self.table.items():
            text += f"{parameter_name}: {parameter.value}, \n"
        print(text)

    def get_parameters(self):
        """
        Get dictionary of all parameters with latest updated values
        Args: parameter_name: Name of the parameter to be returned. If None, all parameters are printed.
        Returns: if parameter_name is None, Dictionary of the form { "parameter_name": parameter_value },
                else parameter_value associated to parameter_name.
        """
        return {parameter_name: parameter.value for parameter_name, parameter in self.table.items()}

    def get_parameter(self, param_name: str):
        """
        Get the value of a specific arameter with latest updated value
        Args: parameter_name: Name of the parameter to be returned. If None, all parameters are printed.
        Returns: if parameter_name is None, Dictionary of the form { "parameter_name": parameter_value },
                else parameter_value associated to parameter_name.
        """
        if param_name not in self.table.keys():
            raise KeyError(f"No parameter named {param_name} in the parameter table.")
        return self.table[param_name].value

    def __getitem__(self, item):
        """
        Returns the QUA variable corresponding to the parameter name.
        """

        if self.table[item].var is not None:
            return self.table[item].var
        else:
            raise ValueError(
                f"No QUA variable found for parameter {item}. Please use "
                f"VideoMode.declare_variables() within QUA program first."
            )

    @property
    def variables(self):
        """List of the QUA variables corresponding to the parameters in the parameter table."""
        try:
            return [self[item] for item in self.table.keys()]
        except KeyError:
            raise KeyError(
                "No QUA variables found for parameters. Please use "
                "VideoMode.declare_variables() within QUA program first."
            )

    def __repr__(self):
        text = ""
        for parameter in self.table.values():
            text += parameter.__repr__()
        return text


class VideoMode:
    """
    This class aims to provide an easy way to update parameters in a QUA program through user input while the
    program is running. It is particularly useful for calibrating parameters dynamically and without deterministic
    of the parameters values. The user can specify
    the parameters to be updated and their initial values in the parameter dictionary called ```param_dict```.
    """

    _default_io1 = 2**31 - 1
    _default_io2 = 0.0

    def __init__(self, qm: QuantumMachine, parameters: Union[Dict, ParameterTable]):
        """
        This class aims to provide an easy way to update parameters in a QUA program through user input while the
        program is running. It is particularly useful for calibrating parameters in real time. The user can specify
        the parameters to be updated and their initial values in the parameter dictionary called ```param_dict```.
        The video mode will then automatically create the corresponding QUA variables and update them through user
        input.

        Parameters dictionary should be of the form:
         ```{"parameter_name": (initial_parameter_value, qua_type) }```
            where ```qua_type``` is the type of the QUA variable to be declared (int, fixed, bool).

        The way this is done is by adding two methods of this class at the beginning of the QUA program declaration:

        - ```declare_variables```: This method will create the QUA variables corresponding to the parameters to be updated.
        - ```load_parameters```: This method will load the updated values for the parameters through IO variables  1 and 2.

        The user can then start the video mode outside the QUA program by calling the ```execute``` method of this class.
        This will start a new parallel thread in charge of updating the
        parameters in the parameter table through user input. The user can stop the video mode by entering 'stop' in
        the terminal. The user can also resume a paused program by entering 'done' again in the terminal.

        Args:
            qm: Quantum Machine object.
            parameters: Dictionary or ParameterTable containing the parameters to be updated and their initial values and types.
        """
        self.qm = qm
        self.job = None
        self._parameter_table = parameters if isinstance(parameters, ParameterTable) else ParameterTable(parameters)
        self.active = True
        self.thread = threading.Thread(target=self.update_parameters)
        self.stop_event = threading.Event()
        self.implemented_commands = (
            "List of implemented commands: \n "
            "get: return the current value of the parameters. \n "
            "stop: quit VideoMode. \n "
            "done: resume program (if pause_program==True). \n "
            "help: display the list of available commands. \n "
            "'param_name'='param_value': set the parameter to the specified value (ex: V1=0.152).\n "
            "'param_name': return the value of the parameter.\n"
        )

        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

    def signal_handler(self, signum, frame):
        """Signal handler for SIGTERM and SIGINT signals."""
        print(f"Received signal {signum}, stopping VideoMode...")
        self.qm.set_io1_value(self._default_io1)
        self.qm.set_io2_value(self._default_io2)
        self.job.halt()
        self.stop_event.set()
        # For shutting down the entire program
        # sys.exit(0)

    def update_parameters(self):
        """Update parameters in the parameter table through user input."""
        while not self.stop_event.is_set() and self.active:
            param_name = input("Enter a command (type help for getting the list of available commands): ")
            messages = param_name.split("=")
            if len(messages) == 1:
                if messages[0] == "stop":
                    self.active = False
                    self.job.halt()
                    break

                elif messages[0] == "done" and self.job is not None:
                    if self.job.is_paused():
                        self.job.resume()

                elif messages[0] == "get":
                    self.parameter_table.print_parameters()

                elif messages[0] == "help":
                    print(self.implemented_commands)

                elif messages[0] in self.parameter_table.table.keys():
                    print(self.parameter_table.table[messages[0]].value)

                else:
                    print(f"Invalid input. {messages[0]} is not a valid command.")

            elif len(messages) == 2:
                param_name = messages[0]
                param_value = messages[1]
                if param_name in self.parameter_table.table.keys():
                    self.qm.set_io1_value(self.parameter_table.table[param_name].index)

                    if isinstance(self.parameter_table.table[param_name].value, list):
                        param_value = param_value.split()
                        if len(param_value) != self.parameter_table.table[param_name].length:
                            print(
                                f"Invalid input. {self.parameter_table[param_name]} should be a list of length "
                                f"{self.parameter_table.table[param_name].length}."
                            )
                        elif param_value[0].isnumeric():
                            try:
                                param_value = list(map(int, param_value))
                            except ValueError:
                                print(
                                    "One of the values could not be cast to int ("
                                    "casting done according to type of first value): "
                                )
                        elif param_value[0].replace(".", "", 1).isdigit():
                            try:
                                param_value = list(map(float, param_value))
                            except ValueError:
                                print(
                                    "One of the values could not be cast to float ("
                                    "casting done based on type of first value): "
                                )
                        elif param_value[0] in ["True", "False"]:
                            try:
                                param_value = list(map(bool, param_value))
                            except ValueError:
                                print(
                                    "One of the values could not be cast to bool (casting done based on type of "
                                    "first value): "
                                )

                        assert all(isinstance(x, type(param_value[0])) for x in param_value), (
                            f"Invalid input. {self.parameter_table[param_name]} should be a list of elements of the "
                            f"same type."
                        )

                        if self.job.is_paused():
                            # If the program is paused before param loading,
                            # update the parameters directly (no wait through done)
                            self.job.resume()
                        for value in param_value:
                            while not (self.job.is_paused()):
                                time.sleep(0.001)
                            self.qm.set_io2_value(value)
                            self.parameter_table.table[param_name].value = value
                            self.job.resume()

                    else:
                        if self.parameter_table.table[param_name].type == int:
                            if not param_value.isnumeric():
                                print(f"Invalid input. {self.parameter_table[param_name]} should be an integer.")
                            else:
                                try:
                                    param_value = int(param_value)
                                except ValueError:
                                    print(
                                        f"Invalid input. {self.parameter_table[param_name]} could not be "
                                        f"converted to int."
                                    )
                                    continue

                        elif self.parameter_table.table[param_name].type == fixed:
                            try:
                                param_value = float(param_value)
                            except ValueError:
                                print(f"Invalid input. {self.parameter_table[param_name]} should be a float.")
                                continue

                        elif self.parameter_table.table[param_name].type == bool:
                            if param_value not in ["True", "False", "0", "1"]:
                                print(f"Invalid input. {self.parameter_table[param_name]} should be a boolean.")
                            elif param_value in ["0", "1"]:
                                param_value = bool(int(param_value))
                            else:
                                try:
                                    param_value = bool(param_value)
                                except ValueError:
                                    print(
                                        f"Invalid input. {self.parameter_table[param_name]} could not be cast to bool"
                                    )
                        else:
                            print(f"Invalid input. {param_value} is not a valid parameter value.")

                        self.qm.set_io2_value(param_value)
                        self.parameter_table.table[param_name].value = param_value

                else:
                    print(f"Invalid input. {param_name} is not a valid parameter name.")

        print("VideoMode stopped.")

    def execute(self, prog: Optional[Program] = None, *execute_args) -> QmJob:
        """Start the video mode.
        Args:
            prog: QUA program to be executed. If None, the video mode will be started without executing a QUA program
            (program assumed to be running)
            execute_args: Arguments to be passed to the execute method of the QuantumMachine for executing the program.
        Returns:
            the QM job

        """
        if self.job is None:
            self.job = self.qm.execute(prog, *execute_args)
            # Reinitialize the IO values
            self.qm.set_io1_value(666)
            self.qm.set_io2_value(0.0)
            time.sleep(1)
            self.job.resume()
        print("start")
        print(self.implemented_commands)
        self.parameter_table.get_parameters()
        self.thread.start()

        return self.job

    @property
    def parameter_table(self):
        """
        ParameterTable containing the parameters to be updated and their initial values.
        """
        return self._parameter_table

    @property
    def variables(self):
        """List of the QUA variables corresponding to the parameters in the parameter table."""
        return self.parameter_table.variables

    def get_parameters(self):
        """
        Get dictionary of parameters with latest updated values
        Returns: Dictionary of the form { "parameter_name": parameter_value }.
        """
        return self.parameter_table.get_parameters()

    def get_parameter(self, param_name: Optional[str] = None):
        """
        Get the value of a parameter with latest updated values
        """
        return self.parameter_table.get_parameter(param_name)

    def load_parameters(self, pause_program=False):
        """
        QUA Macro to be called within QUA program to retrieve updated values for the parameters through IO 1 and IO2.
        Args:
        pause_program: Boolean indicating whether the program should be paused while waiting for user input.
        """
        self.parameter_table.load_parameters(pause_program)

    def declare_variables(self):
        """
        QUA Macro to create the QUA variables associated with the parameter table.
        :returns : The list of declared QUA variables or the declared variable if there is only one item.
        """
        return self.parameter_table.declare_variables()

    def __getitem__(self, item):
        """
        Returns the QUA variable corresponding to the parameter name.
        """
        return self._parameter_table[item]



from typing import List, Union

from .instrument_channel import (
    InstrumentChannelOctaveInput,
    InstrumentChannelOctaveOutput,
    InstrumentChannelOctaveDigitalInput,
    InstrumentChannelExternalMixerInput,
    InstrumentChannelExternalMixerOutput,
    InstrumentChannelExternalMixerDigitalInput,
)
from .instrument_channels import *
from .constants import *


class Instruments:
    """
    Class to add the static information about which QM instruments will be used
    in an experimental setup. Upon adding an instrument, its available channels
    will be enumerated and added individually to a stack of free channels.
    """

    def __init__(self):
        self.used_channels = InstrumentChannels()
        self.available_channels = InstrumentChannels()

    def add_external_mixer(self, indices: Union[List[int], int]):
        """
        Add an external mixer, which is defined abstractly as a combined, IQ-upconverter and
        IQ-downconverter.

        `indices` (List[int] | int): Can be one or more indices for one or more external mixers.
        """
        if isinstance(indices, int):
            indices = [indices]

        for index in indices:
            channel = InstrumentChannelExternalMixerInput(con=index, port=1)
            self.available_channels.add(channel)

            channel = InstrumentChannelExternalMixerOutput(con=index, port=1)
            self.available_channels.add(channel)

            channel = InstrumentChannelExternalMixerDigitalInput(con=index, port=1)
            self.available_channels.add(channel)

    def add_octave(self, indices: Union[List[int], int]):
        if isinstance(indices, int):
            indices = [indices]

        for index in indices:
            for port in range(1, NUM_OCTAVE_INPUT_PORTS + 1):
                channel = InstrumentChannelOctaveInput(con=index, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_OCTAVE_OUTPUT_PORTS + 1):
                channel = InstrumentChannelOctaveOutput(con=index, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_OCTAVE_DIGITAL_INPUT_PORTS + 1):
                channel = InstrumentChannelOctaveDigitalInput(con=index, port=port)
                self.available_channels.add(channel)

    def add_lf_fem(self, controller: int, slots: Union[List[int], int]):
        if isinstance(slots, int):
            slots = [slots]

        for slot in slots:
            for port in range(1, NUM_LF_FEM_INPUT_PORTS + 1):
                channel = InstrumentChannelLfFemInput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_LF_FEM_OUTPUT_PORTS + 1):
                channel = InstrumentChannelLfFemOutput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_LF_FEM_DIGITAL_OUTPUT_PORTS + 1):
                channel = InstrumentChannelLfFemDigitalOutput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

    def add_mw_fem(self, controller: int, slots: Union[List[int], int]):
        if isinstance(slots, int):
            slots = [slots]

        for slot in slots:
            for port in range(1, NUM_MW_FEM_INPUT_PORTS + 1):
                channel = InstrumentChannelMwFemInput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_MW_FEM_OUTPUT_PORTS + 1):
                channel = InstrumentChannelMwFemOutput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_MW_FEM_OUTPUT_PORTS + 1):
                channel = InstrumentChannelMwFemDigitalOutput(con=controller, slot=slot, port=port)
                self.available_channels.add(channel)

    def add_opx_plus(self, controllers: Union[List[int], int]):
        if isinstance(controllers, int):
            controllers = [controllers]

        for controller in controllers:
            for port in range(1, NUM_OPX_PLUS_INPUT_PORTS + 1):
                channel = InstrumentChannelOpxPlusInput(con=controller, port=port)
                self.available_channels.add(channel)

            for port in range(1, NUM_OPX_PLUS_OUTPUT_PORTS + 1):
                channel = InstrumentChannelOpxPlusOutput(con=controller, port=port)
                self.available_channels.add(channel)

            # add odd first for octave connectivity (top row is more convenient)
            for port in range(1, NUM_OPX_PLUS_DIGITAL_OUTPUT_PORTS + 1, 2):
                channel = InstrumentChannelOpxPlusDigitalOutput(con=controller, port=port)
                self.available_channels.add(channel)

            # then add evens
            for port in range(2, NUM_OPX_PLUS_DIGITAL_OUTPUT_PORTS + 1, 2):
                channel = InstrumentChannelOpxPlusDigitalOutput(con=controller, port=port)
                self.available_channels.add(channel)



from abc import ABC, abstractmethod
import dataclasses
from time import sleep
from typing import List, Any, Dict
from functools import wraps
from qm import QmJob
from qm.program import Program
from qm.qua import declare_stream, save, pause
from qm.exceptions import QmQuaException
from qm import QuantumMachine
from qm.qua._dsl import _ResultSource, _Variable, align, _get_root_program_scope

__all__ = ["ProgramAddon", "callable_from_qua"]


@dataclasses.dataclass
class QuaCallableArgument:
    tag: str
    stream: _ResultSource


class QuaCallable:
    def __init__(
        self,
        fn: callable,
        qua_callable_stream: _ResultSource,
        qua_callable_id: int,
        args: List[Any],
        kwargs: Dict[str, Any],
    ):
        self._fn = fn
        self._qua_callable_id = qua_callable_id
        self._streams: Dict[str, _ResultSource] = {}
        self._declare(qua_callable_stream, args, kwargs)
        self._last_arg_fetch: Dict[str, int] = {}

    def _declare(
        self,
        qua_callable_stream: _ResultSource,
        args: List[Any],
        kwargs: Dict[str, Any],
    ):
        align()
        self._args = [self._convert_to_qua_arg(f"__pos__{i}", arg) for i, arg in enumerate(args)]
        self._kwargs = {name: self._convert_to_qua_arg(name, arg) for name, arg in kwargs.items()}
        save(self._qua_callable_id, qua_callable_stream)
        pause()
        align()

    def _convert_to_qua_arg(self, arg_id: str, arg):
        if isinstance(arg, _Variable):
            tag = f"__qua_callable__{self._qua_callable_id}__{arg_id}"
            stream = declare_stream()
            save(arg, stream)
            self._streams[tag] = stream
            return QuaCallableArgument(tag, stream)
        else:
            return arg

    def do_stream_processing(self):
        for tag, stream in self._streams.items():
            stream.with_timestamps().save(tag)

    def _convert_to_value_arg(self, job: QmJob, arg):
        if not isinstance(arg, QuaCallableArgument):
            return arg
        while True:
            value = job.result_handles.get(arg.tag).fetch_all()
            if value is not None and (arg.tag not in self._last_arg_fetch or self._last_arg_fetch[arg.tag] != value[1]):
                self._last_arg_fetch[arg.tag] = value[1]
                return value[0]
            sleep(0.01)

    def run(self, job: QmJob):
        args = [self._convert_to_value_arg(job, arg) for arg in self._args]
        kwargs = {name: self._convert_to_value_arg(job, arg) for name, arg in self._kwargs.items()}
        return self._fn(*args, **kwargs)


class ProgramAddon(ABC):
    @abstractmethod
    def enter_program(self, program: Program): ...  # noqa: E704

    @abstractmethod
    def exit_program(self, exc_type, exc_val, exc_tb): ...  # noqa: E704

    @abstractmethod
    def execute_program(self, program: Program, quantum_machine: QuantumMachine): ...  # noqa: E704


class QuaCallableEventManager(ProgramAddon):
    def __init__(self):
        """Framework allowing the user to call Python functions directly from the core of a QUA program.
        The Python function can be wrapped under a Python or QUA for loop and variables can be passed to and from this function (Python or QUA variables).
        This can be used to update the parameters of external instruments such as the Octave LO frequency and power, or the QDAC channel voltages for instance.
        It can also be used to post-process data in Python (machine learning, fitting, ...) and update the running QUA program seamlessly.
        It can also be used to quickly print the values of QUA variables for debugging purposes.
        """
        self._qua_callables: List[QuaCallable] = []
        self._qua_callable_stream = None
        self._last_qua_callable = -1
        self.callables: List[callable] = []  # TODO maybe merge with QuaCallable?
        self._qm = None  # TODO needed to transfer data from python to QUA with iovalues

        try:
            # Check if we are already inside a program
            _get_root_program_scope()
            self.declare_all()
        except QmQuaException:
            ...

    def enter_program(self, program: Program):
        self.declare_all()

    def exit_program(self, exc_type, exc_val, exc_tb):
        self.do_stream_processing()

    def execute_program(self, program: Program, quantum_machine: QuantumMachine):
        """
        Executes the program using the given quantum machine and set of local run functions called in the QUA program.

        Args:
            program: The program object.
            quantum_machine: The quantum machine object.
        """
        self._qm = quantum_machine  # TODO needed to transfer data from python to QUA with iovalues
        job = quantum_machine.get_running_job()
        result_handles = job.result_handles

        while result_handles.is_processing():
            self.attempt_qua_callable(job)

            for func in self.callables:
                func(result_handles)

    def register_qua_callable(self, fn: callable, *args, **kwargs):
        lr = QuaCallable(fn, self._qua_callable_stream, len(self._qua_callables), list(args), kwargs)
        self._qua_callables.append(lr)

    def declare_all(self):
        self._qua_callable_stream = declare_stream()

    def do_stream_processing(self):
        for qua_callable in self._qua_callables:
            qua_callable.do_stream_processing()
        if len(self._qua_callables) > 0:
            self._qua_callable_stream.with_timestamps().save("__qua_callable")

    def attempt_qua_callable(self, job: QmJob):
        if not job.is_paused():
            return
        qua_callable_id = job.result_handles.get("__qua_callable").fetch_all()
        if qua_callable_id is not None and qua_callable_id[1] != self._last_qua_callable:
            self._last_qua_callable = qua_callable_id[1]
            out = self._qua_callables[qua_callable_id[0]].run(
                job
            )  # TODO needed to transfer data from python to QUA with iovalues
            if out is not None:
                if isinstance(out, list) or isinstance(out, tuple):
                    if len(out) == 2:
                        self._qm.set_io1_value(out[0])
                        self._qm.set_io2_value(out[1])
                else:
                    self._qm.set_io1_value(out[0])
            job.resume()


def callable_from_qua(func: callable):
    """Decorator that allows the call of functions, to be executed in locally Python (on the control PC), directly from a QUA program.

    :param func: The function to be executed from a QUA program.
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            program = _get_root_program_scope()._program
        except IndexError:
            return func(*args, **kwargs)

        if not hasattr(program, "addons"):
            raise RuntimeError(
                "Cannot execute callable_from_qua function in a program until program is patched. "
                "Please first run qualang_tools.callable_from_qua.patch_qua_program_addons()"
            )

        if "callable_from_qua" not in program.addons:
            program.addons["callable_from_qua"] = QuaCallableEventManager()

        event_manager = program.addons["callable_from_qua"]
        # active_program_manager = QuaCallableEventManager._active_program_manager
        event_manager.register_qua_callable(func, *args, **kwargs)

    return wrapper



from typing import Dict, List, Union

from .channel_spec import ChannelSpec
from .element import Element, ElementId, QubitReference, QubitPairReference
from .types import QubitsType, QubitPairsType
from .wiring_spec import WiringSpec, WiringFrequency, WiringIOType, WiringLineType


class ConnectivityBase:
    """
    This class stores placeholders for the quantum elements which will be used
    in a setup, as well as the wiring specification for each of those elements.
    This includes at what frequency the element will be driven, whether it
    requires input/output or both lines, if it is required on a particular
    module or FEM slot, and what high-level component will be manipulated.
    """

    def __init__(self):
        self.elements: Dict[ElementId, Element] = {}
        self.specs: List[WiringSpec] = []

    def add_wiring_spec(
        self,
        frequency: WiringFrequency,
        io_type: WiringIOType,
        line_type: Union[WiringLineType, str],
        triggered: bool,
        constraints: ChannelSpec,
        elements: List[Element],
        shared_line: bool = False,
    ):
        specs = []
        for element in elements:
            if element.id not in self.elements:
                self._add_element(element)

        if shared_line:
            spec = WiringSpec(frequency, io_type, line_type, triggered, constraints, elements)
            specs.append(spec)
        else:
            for element in elements:
                spec = WiringSpec(frequency, io_type, line_type, triggered, constraints, element)
                specs.append(spec)
        self.specs.extend(specs)

        return specs

    def _make_qubit_elements(self, qubits: QubitsType):
        if not isinstance(qubits, list):
            qubits = [qubits]

        elements = []
        for qubit in qubits:
            id = QubitReference(qubit)
            if id not in self.elements:
                self.elements[id] = Element(id)
            elements.append(self.elements[id])

        return elements

    def _make_qubit_pair_elements(self, qubit_pairs: QubitPairsType):
        if not isinstance(qubit_pairs, list):
            qubit_pairs = [qubit_pairs]

        elements = []
        for qubit_pair in qubit_pairs:
            id = QubitPairReference(*qubit_pair)
            if id not in self.elements:
                self.elements[id] = Element(id)
            elements.append(self.elements[id])

        return elements

    def _add_elements(self, elements: List[Element]):
        for element in elements:
            self._add_element(element)

    def _add_element(self, element: Element):
        self.elements[element.id] = element



__all__ = [
    "simulator",
    "bakery",
    "config",
    "control_panel",
    "loops",
    "results",
    "plot",
    "analysis",
    "macros",
    "multi_user",
    "units",
    "external_frameworks",
    "callable_from_qua",
    "wirer",
]



from ._cirq import import_cirq
from .RBBaker import RBBaker
from .RBResult import RBResult
from .gates import GateGenerator, gate_db, tableau_from_cirq
from .simple_tableau import SimpleTableau
from .util import run_in_thread, pbar
from .verification.command_registry import (
    CommandRegistry,
    decorate_single_qubit_generator_with_command_recording,
    decorate_two_qubit_gate_generator_with_command_recording,
)
from .verification.sequence_tracker import SequenceTracker

from pathlib import Path
from typing import Callable, List, Literal, Dict, Tuple, Optional, Union

import numpy as np
from qm import QuantumMachinesManager
from qm.jobs.running_qm_job import RunningQmJob
from qm.qua import *

from qualang_tools.bakery.bakery import Baking

cirq = import_cirq()


class TwoQubitRb:
    _buffer_length = 4096

    def __init__(
        self,
        config: dict,
        single_qubit_gate_generator: Callable[[Baking, int, float, float, float], None],
        two_qubit_gate_generators: Dict[Literal["sqr_iSWAP", "CNOT", "CZ"], Callable[[Baking, int, int], None]],
        prep_func: Callable[[], None],
        measure_func: Callable[[], Tuple],
        verify_generation: bool = False,
        interleaving_gate: Optional[List[cirq.GateOperation]] = None,
    ):
        """
        A class for running two qubit randomized benchmarking experiments.

        This class is used to generate the experiment configuration and run the experiment.
        The experiment is run by calling the run method.

        Gate generation is performed using the `Baking`[https://github.com/qua-platform/py-qua-tools/blob/main/qualang_tools/bakery/README.md] class.
        This class adds to QUA the ability to generate arbitrary waveforms ("baked waveforms") using syntax similar to QUA.

        Args:
            config: A QUA configuration containing the configuration for the experiment.

            single_qubit_gate_generator: A callable used to generate a single qubit gate using a signature similar to `phasedXZ`[https://quantumai.google/reference/python/cirq/PhasedXZGate].
                This is done using the baking object (see above).
                Note that this allows us to execute every type of single qubit gate.
                Callable arguments:
                    baking: The baking object.
                    qubit: The qubit number.
                    x: The x rotation exponent.
                    z: The z rotation exponent.
                    a: the axis phase exponent.

            two_qubit_gate_generators: A dictionary mapping one or more two qubit gate names to callables used to generate those gates.
                This is done using the baking object (see above).
                Callable arguments:
                    baking: The baking object.
                    qubit1: The first qubit number.
                    qubit2: The second qubit number.
                This callable should generate a two qubit gate.

            prep_func: A callable used to reset the qubits to the |00> state. This function does not use the baking object, and is a proper QUA code macro.
                Callable arguments: None

            measure_func: A callable used to measure the qubits. This function does not use the baking object, and is a proper QUA code macro.
                Callable[[], Tuple[_Expression, _Expression]]: A tuple containing the measured values of the two qubits as Qua expressions.
                The expression must evaluate to a boolean value. False means |0>, True means |1>. The MSB is the first qubit.

            verify_generation: A boolean indicating whether to verify the generated sequences. Not be used in production, as it is very slow.

            interleaving_gate: Interleaved gate represented as list of cirq GateOperation
        """
        for i, qe in config["elements"].items():
            if "operations" not in qe:
                qe["operations"] = {}

        self._command_registry = CommandRegistry()
        self._sequence_tracker = SequenceTracker(command_registry=self._command_registry)

        single_qubit_gate_generator = decorate_single_qubit_generator_with_command_recording(
            single_qubit_gate_generator, self._command_registry
        )
        two_qubit_gate_generators = decorate_two_qubit_gate_generator_with_command_recording(
            two_qubit_gate_generators, self._command_registry
        )
        self._rb_baker = RBBaker(
            config, single_qubit_gate_generator, two_qubit_gate_generators, interleaving_gate, self._command_registry
        )

        self._interleaving_gate = interleaving_gate
        self._interleaving_tableau = tableau_from_cirq(interleaving_gate) if interleaving_gate is not None else None
        self._config = self._rb_baker.bake()
        self._symplectic_generator = GateGenerator(set(two_qubit_gate_generators.keys()))
        self._prep_func = prep_func
        self._measure_func = measure_func
        self._verify_generation = verify_generation

    def convert_sequence_to_cirq(self, sequence: List[int]) -> List[cirq.GateOperation]:
        gates = []
        for cmd_id in sequence:
            gates.extend(self._rb_baker.gates_from_cmd_id(cmd_id))
        return gates

    def _verify_rb_sequence(self, gate_ids, final_tableau: SimpleTableau):
        if final_tableau != SimpleTableau(np.eye(4), [0, 0, 0, 0]):
            raise RuntimeError("Verification of RB sequence failed")
        gates = []
        for gate_id in gate_ids:
            if gate_id == gate_db.get_interleaving_gate():
                gates.extend(self._interleaving_gate)
            else:
                gates.extend(self._symplectic_generator.generate(gate_id))

        unitary = cirq.Circuit(gates).unitary()
        fixed_phase_unitary = np.conj(np.trace(unitary) / 4) * unitary
        if np.linalg.norm(fixed_phase_unitary - np.eye(4)) > 1e-12:
            raise RuntimeError("Verification of RB sequence failed")

    def _gen_rb_sequence(self, depth):
        gate_ids = []
        tableau = SimpleTableau(np.eye(4), [0, 0, 0, 0])
        for i in range(depth):
            symplectic = gate_db.rand_symplectic()
            pauli = gate_db.rand_pauli()
            gate_ids.append(symplectic)
            gate_ids.append(pauli)

            tableau = tableau.then(gate_db.get_tableau(symplectic)).then(gate_db.get_tableau(pauli))

            if self._interleaving_tableau is not None:
                gate_ids.append(gate_db.get_interleaving_gate())
                tableau = tableau.then(self._interleaving_tableau)

        inv_tableau = tableau.inverse()
        inv_id = gate_db.find_symplectic_gate_id_by_tableau_g(inv_tableau)
        after_inv_tableau = tableau.then(gate_db.get_tableau(inv_id))

        pauli = gate_db.find_pauli_gate_id_by_tableau_alpha(after_inv_tableau)

        gate_ids.append(inv_id)
        gate_ids.append(pauli)

        if self._verify_generation:
            final_tableau = after_inv_tableau.then(gate_db.get_tableau(pauli))
            self._verify_rb_sequence(gate_ids, final_tableau)

        return gate_ids

    def _gen_qua_program(self, sequence_depths: list[int], num_repeats: int, num_averages: int, unsafe: bool):
        with program() as prog:
            sequence_depth = declare(int)
            repeat = declare(int)
            n_avg = declare(int)
            state = declare(int)
            length = declare(int)
            progress = declare(int)
            progress_os = declare_stream()
            state_os = declare_stream()
            gates_len_is = declare_input_stream(int, name="__gates_len_is__", size=1)
            gates_is = {
                qe: declare_input_stream(int, name=f"{self._input_stream_name(qe)}_is", size=self._buffer_length)
                for qe in self._rb_baker.all_elements
            }

            assign(progress, 0)
            with for_each_(sequence_depth, sequence_depths):
                with for_(repeat, 0, repeat < num_repeats, repeat + 1):
                    assign(progress, progress + 1)
                    save(progress, progress_os)
                    advance_input_stream(gates_len_is)
                    for gate_is in gates_is.values():
                        advance_input_stream(gate_is)
                    assign(length, gates_len_is[0])
                    with for_(n_avg, 0, n_avg < num_averages, n_avg + 1):
                        self._prep_func()
                        self._rb_baker.run(gates_is, length, unsafe=unsafe)
                        out1, out2 = self._measure_func()
                        assign(state, (Cast.to_int(out2) << 1) + Cast.to_int(out1))
                        save(state, state_os)

            with stream_processing():
                state_os.buffer(len(sequence_depths), num_repeats, num_averages).save("state")
                progress_os.save("progress")
        return prog

    def _input_stream_name(self, element: str):
        return element.replace(".", "__dot__")

    def _decode_sequence_for_element(self, element: str, seq: list):
        seq = [self._rb_baker.decode(i, element) for i in seq]
        if len(seq) > self._buffer_length:
            RuntimeError("Buffer is too small")
        return seq + [0] * (self._buffer_length - len(seq))

    @run_in_thread
    def _insert_all_input_stream(
        self,
        job: RunningQmJob,
        sequence_depths: List[int],
        num_repeats: int,
        callback: Optional[Callable[[List[int]], None]] = None,
    ):
        for sequence_depth in sequence_depths:
            for repeat in range(num_repeats):
                sequence = self._gen_rb_sequence(sequence_depth)
                if self._sequence_tracker is not None:
                    self._sequence_tracker.make_sequence(sequence)
                job.insert_input_stream("__gates_len_is__", len(sequence))
                for qe in self._rb_baker.all_elements:
                    job.insert_input_stream(f"{qe}_is", self._decode_sequence_for_element(qe, sequence))

                if callback is not None:
                    callback(sequence)

    def run(
        self,
        qmm: QuantumMachinesManager,
        circuit_depths: List[int],
        num_circuits_per_depth: int,
        num_shots_per_circuit: int,
        unsafe: bool = False,
        **kwargs,
    ):
        """
        Runs the randomized benchmarking experiment. The experiment is sweep over Clifford circuits with varying depths.
        For every depth, we generate a number of random circuits and run them. The number of different circuits is determined by
        the num_circuits_per_depth parameter. The number of shots per individual circuit is determined by the num_averages parameter.

        Args:
            qmm (QuantumMachinesManager): The Quantum Machines Manager object which is used to run the experiment.
            circuit_depths (List[int]): A list of the number of Cliffords per circuit (not including inverse).
            num_circuits_per_depth (int): The number of different circuit randomizations per depth.
            num_shots_per_circuit (int): The number of shots per particular circuit.
            unsafe (bool): Refers to the option of compiling a QUA switch-case "safely", which
                           guarantees correct behaviour but can lead to gaps, or "unsafely",
                           which reduces gaps but can cause unwanted behaviour. Note: as of
                           QOP 3.2.3, there seems to be an issue with "unsafe" compilation.

        """
        prog = self._gen_qua_program(circuit_depths, num_circuits_per_depth, num_shots_per_circuit, unsafe)

        qm = qmm.open_qm(self._config)
        job = qm.execute(prog)

        gen_sequence_callback = kwargs["gen_sequence_callback"] if "gen_sequence_callback" in kwargs else None
        self._insert_all_input_stream(job, circuit_depths, num_circuits_per_depth, gen_sequence_callback)

        full_progress = len(circuit_depths) * num_circuits_per_depth
        pbar(job.result_handles, full_progress, "progress")
        job.result_handles.wait_for_all_values()

        return RBResult(
            circuit_depths=circuit_depths,
            num_repeats=num_circuits_per_depth,
            num_averages=num_shots_per_circuit,
            state=job.result_handles.get("state").fetch_all(),
        )

    def print_command_mapping(self):
        """
        Prints the mapping of Command ID index, which is understood by the
        input stream, into single-qubit and two-qubit gates.
        """
        self._command_registry.print_commands()

    def print_sequences(self):
        """
        Prints a break-down of all gates/commands which were played in
        each random sequence.
        """
        self._sequence_tracker.print_sequences()

    def save_command_mapping_to_file(self, path: Union[str, Path]):
        """
        Saves a text file containing the mapping of Command ID index, which
        is understood by the input stream, into single-qubit and two-qubit gates.
        """
        self._command_registry.save_to_file(path)

    def save_sequences_to_file(self, path: Union[str, Path]):
        """
        Save a text file of the break-down of all gates/commands which
        were played in each random sequence
        """
        self._sequence_tracker.save_to_file(path)

    def verify_sequences(self):
        """
        Simulates the application of all random sequences on the |00> state
        to ensure that they recover the qubit to |00> correctly.

        Note: You should only call this function *after* self.run().
        """
        self._sequence_tracker.verify_sequences()



"""Tools to help to handle plots from QUA programs.

Content:
    - interrupt_on_close: Allows to interrupt the execution and free the console when closing the live-plotting figure.
"""

from typing import List, Tuple
import numpy as np
from scipy import signal
import plotly.graph_objects as go
from matplotlib import pyplot as plt
from qm import QmJob
from qualang_tools.plot.fitting import *


def plot_demodulated_data_2d(
    x,
    y,
    I,
    Q,
    x_label: str = None,
    y_label: str = None,
    title: str = None,
    amp_and_phase: bool = True,
    fig=None,
    plot_options: dict = None,
):
    """
    Plots 2D maps of either 'I' and 'Q' or the corresponding amplitude and phase in a single figure.

    :param x: 1D numpy array representing the x-axis.
    :param y: 1D numpy array representing the y-axis.
    :param I: 2D numpy array representing the 'I' quadrature.
    :param Q: 2D numpy array representing the 'Q' quadrature.
    :param x_label: name for the x-axis label.
    :param y_label: name for the y-axis label.
    :param title: title of the plot.
    :param amp_and_phase: if True, plots the amplitude [np.sqrt(I**2 + Q**2)] and phase [signal.detrend(np.unwrap(np.angle(I + 1j * Q)))] instead of I and Q.
    :param fig: a matplotlib figure. If `none` (default), will create a new one.
    :param plot_options: dictionary containing various plotting options. Defaults are ['fontsize': 14, 'cmap': `magma`, 'figsize': None].
    :return: the matplotlib figure object.
    """
    _plot_options = {
        "fontsize": 14,
        "cmap": "magma",
        "figsize": None,
    }
    if plot_options is not None:
        for key in plot_options:
            if key in _plot_options.keys():
                _plot_options[key] = plot_options[key]
            else:
                raise ValueError(
                    f"The key {key} in 'plot_options' doesn't exists. The available options are {[option for option in _plot_options.keys()]}"
                )
    if amp_and_phase:
        z1 = np.sqrt(I**2 + Q**2)
        z2 = signal.detrend(np.unwrap(np.angle(I + 1j * Q)))
        z1_label = "Amplitude [a.u.]"
        z2_label = "Phase [rad]"
    else:
        z1 = I
        z2 = Q
        z1_label = "I [a.u.]"
        z2_label = "Q [a.u.]"

    if fig is None:
        fig = plt.figure(figsize=_plot_options["figsize"])
    plt.suptitle(title, fontsize=_plot_options["fontsize"] + 2)
    ax1 = plt.subplot(211)
    plt.cla()
    plt.title(z1_label, fontsize=_plot_options["fontsize"])
    plt.pcolor(x, y, z1, cmap=_plot_options["cmap"])
    plt.ylabel(y_label, fontsize=_plot_options["fontsize"])
    plt.xticks(fontsize=_plot_options["fontsize"])
    plt.yticks(fontsize=_plot_options["fontsize"])
    plt.colorbar()
    plt.subplot(212, sharex=ax1)
    plt.cla()
    plt.title(z2_label, fontsize=_plot_options["fontsize"])
    plt.pcolor(x, y, z2, cmap=_plot_options["cmap"])
    plt.xlabel(x_label, fontsize=_plot_options["fontsize"])
    plt.ylabel(y_label, fontsize=_plot_options["fontsize"])
    plt.xticks(fontsize=_plot_options["fontsize"])
    plt.yticks(fontsize=_plot_options["fontsize"])
    plt.colorbar()
    plt.pause(0.01)
    plt.tight_layout()
    return fig


def plot_demodulated_data_1d(
    x: np.ndarray,
    I: np.ndarray,
    Q: np.ndarray,
    x_label: str = None,
    title: str = None,
    amp_and_phase: bool = True,
    fig=None,
    plot_options: dict = None,
    fit=None,
):
    """
    Plots 1D graphs of either 'I' and 'Q' or the corresponding amplitude and phase in a single figure.

    :param x: 1D numpy array representing the x-axis.
    :param I: 1D numpy array representing the 'I' quadrature.
    :param Q: 1D numpy array representing the 'Q' quadrature.
    :param x_label: name for the x-axis label.
    :param title: title of the plot.
    :param amp_and_phase: if `True` (default value), plots the amplitude [np.sqrt(I**2 + Q**2)] and phase [signal.detrend(np.unwrap(np.angle(I + 1j * Q)))] instead of I and Q.
    :param fig: a matplotlib figure. If `none` (default), will create a new one.
    :param plot_options: dictionary containing various plotting options. Defaults are ['fontsize': 14, 'color': 'b', 'marker': 'o', 'linewidth': 0, 'figsize': None].
    :param fit: Name of the fitting function to be used: 'rabi', 'ramsey', 'T1', 'transmission_resonator_spectroscopy', 'reflection_resonator_spectroscopy'.
    :return: the matplotlib figure object.
    """
    _plot_options = {
        "fontsize": 14,
        "color": "b",
        "marker": "o",
        "linewidth": 0,
        "figsize": None,
    }
    if plot_options is not None:
        for key in plot_options:
            if key in _plot_options.keys():
                _plot_options[key] = plot_options[key]
            else:
                raise ValueError(
                    f"The key '{key}' in 'plot_options' doesn't exists. The available options are {[option for option in _plot_options.keys()]}"
                )

    if amp_and_phase:
        y1 = np.sqrt(I**2 + Q**2)
        y2 = signal.detrend(np.unwrap(np.angle(I + 1j * Q)))
        y1_label = "Amplitude [a.u.]"
        y2_label = "Phase [rad]"
    else:
        y1 = I
        y2 = Q
        y1_label = "I [a.u.]"
        y2_label = "Q [a.u.]"

    if fig is None:
        fig = plt.figure(figsize=_plot_options["figsize"])
    plt.suptitle(title, fontsize=_plot_options["fontsize"] + 2)
    ax1 = plt.subplot(211)
    plt.cla()
    plt.plot(
        x,
        y1,
        color=_plot_options["color"],
        marker=_plot_options["marker"],
        linewidth=_plot_options["linewidth"],
    )
    try:
        if fit == "rabi":
            Fit.rabi(x, y1, plot=True)
        elif fit == "ramsey":
            Fit.ramsey(x, y1, plot=True)
        elif fit == "T1":
            Fit.T1(x, y1, plot=True)
        elif fit == "transmission_resonator_spectroscopy":
            Fit.transmission_resonator_spectroscopy(x, y1, plot=True)
        elif fit == "reflection_resonator_spectroscopy":
            Fit.reflection_resonator_spectroscopy(x, y1, plot=True)
    except (Exception,):
        pass
    plt.ylabel(y1_label, fontsize=_plot_options["fontsize"])
    plt.xticks(fontsize=_plot_options["fontsize"])
    plt.yticks(fontsize=_plot_options["fontsize"])
    plt.subplot(212, sharex=ax1)
    plt.cla()
    plt.plot(
        x,
        y2,
        color=_plot_options["color"],
        marker=_plot_options["marker"],
        linewidth=_plot_options["linewidth"],
    )
    try:
        if fit == "rabi":
            Fit.rabi(x, y2, plot=True)
        elif fit == "ramsey":
            Fit.ramsey(x, y2, plot=True)
        elif fit == "T1":
            Fit.T1(x, y2, plot=True)
        elif fit == "transmission_resonator_spectroscopy":
            Fit.transmission_resonator_spectroscopy(x, y2, plot=True)
        elif fit == "reflection_resonator_spectroscopy":
            Fit.reflection_resonator_spectroscopy(x, y2, plot=True)
    except (Exception,):
        pass
    plt.xlabel(x_label, fontsize=_plot_options["fontsize"])
    plt.ylabel(y2_label, fontsize=_plot_options["fontsize"])
    plt.xticks(fontsize=_plot_options["fontsize"])
    plt.yticks(fontsize=_plot_options["fontsize"])
    plt.pause(0.01)
    plt.tight_layout()
    return fig, fit


def get_simulated_samples_by_element(element_name: str, job: QmJob, config: dict) -> Tuple:
    """
    Extract the simulated samples (analog and digital) corresponding to a given element.

    :param element_name: name of the element corresponding to the samples to extract. Must be defined in the config.
    :param job: The simulated QmJob from which the samples will be extracted.
    :param config: The config file used to create the job.
    :return: Two 1D lists containing the simulated analog and digital samples for the specified element.
    """
    element = config["elements"][element_name]
    job.result_handles.wait_for_all_values()
    sample_struct = job.get_simulated_samples()
    # Analog waveforms
    if "mixInputs" in element:
        port_i = element["mixInputs"]["I"]
        port_q = element["mixInputs"]["Q"]
        analog_samples = (
            sample_struct.__dict__[port_i[0]].analog[str(port_i[1])]
            + 1j * sample_struct.__dict__[port_q[0]].analog[str(port_q[1])]
        )
    elif "singleInput" in element:
        port = element["singleInput"]["port"]
        analog_samples = sample_struct.__dict__[port[0]].analog[str(port[1])]
    else:
        analog_samples = None
    # Digital waveforms
    digital_samples = []
    if "digitalInputs" in element:

        for key in element["digitalInputs"].keys():
            port = element["digitalInputs"][key]["port"]
            digital_samples.append(sample_struct.__dict__[port[0]].digital[str(port[1])].astype(int))
    else:
        digital_samples.append(None)

    return analog_samples, digital_samples


def plot_simulator_output(
    job: QmJob,
    config: dict,
    duration_ns: int,
    plot_axes: List[List[str]] = (()),
    qua_program=None,
):
    """
    Generate a 'plotly' plot of simulator output by elements. Note that this works only if the elements don't share the same ports.

    :param job: The simulated QmJob to plot.
    :param config: The config file used to create the job.
    :param duration_ns: the duration to plot in nanosecond.
    :param plot_axes: (Either this or 'program' have to be provided) a list of lists of elements (ex: [["qubit0", "qubit1"], ["resonator0", "resonator1"]]). Will open multiple axes, one for each list.
    :param qua_program: (Either this or 'plot_axes' have to be provided) All the elements will be plotted in their own axis.
    :return: the generated 'plotly' figure.
    """
    time_vec = np.linspace(0, duration_ns - 1, duration_ns)
    samples_struct = []
    digital_samples_struct = []
    if plot_axes == (()):
        if qua_program is not None:
            # Find the elements used in the program
            el_list = []
            for statement in qua_program.__dict__["_program"].script.body.statements:
                if statement.play.qe.name not in el_list and statement.play.qe.name != "":
                    el_list.append(statement.play.qe.name)
        else:
            raise Exception(
                "If plot_axis is not provided, then the qua_program must be specified as kwargs using qua_program=."
            )
        plot_axes = [[element] for element in el_list]
    for plot_axis in plot_axes:
        samples_struct.append([get_simulated_samples_by_element(pa, job, config)[0] for pa in plot_axis])
        digital_samples_struct.append([get_simulated_samples_by_element(pa, job, config)[1] for pa in plot_axis])

    fig = go.Figure().set_subplots(rows=len(plot_axes), cols=1, shared_xaxes=True)

    plot_item = None
    for i, plot_axis in enumerate(plot_axes):
        for j, plot_item in enumerate(plot_axis):
            if samples_struct[i][j] is not None:
                if isinstance(samples_struct[i][j][0], float):
                    fig.add_trace(
                        go.Scatter(x=time_vec, y=samples_struct[i][j], name=plot_item + " [V]"),
                        row=i + 1,
                        col=1,
                    )
                else:
                    fig.add_trace(
                        go.Scatter(
                            x=time_vec,
                            y=samples_struct[i][j].real,
                            name=plot_item + "- I [V]",
                        ),
                        row=i + 1,
                        col=1,
                    )
                    fig.add_trace(
                        go.Scatter(
                            x=time_vec,
                            y=samples_struct[i][j].imag,
                            name=plot_item + "- Q [V]",
                        ),
                        row=i + 1,
                        col=1,
                    )

            for k in range(len(digital_samples_struct[i][j])):
                if digital_samples_struct[i][j][k] is not None:
                    fig.add_trace(
                        go.Scatter(
                            x=time_vec,
                            y=digital_samples_struct[i][j][k],
                            name=plot_item
                            + " - "
                            + list(config["elements"][plot_item]["digitalInputs"].keys())[k]
                            + " [3.3V]",
                        ),
                        row=i + 1,
                        col=1,
                    )
        fig.get_subplot(i + 1, 1).yaxis.title.text = plot_item

    fig.get_subplot(len(plot_axes), 1).xaxis.title.text = "time [ns]"
    fig.update_layout(title="Simulated waveforms")
    return fig


def interrupt_on_close(figure, current_job):
    """
    Allows to interrupt the execution and free the console when closing the live-plotting figure.

    :param figure: the python figure instance corresponding to the live-plotting figure.
    :param current_job: a ``QmJob`` object (see QM Job API) corresponding to the running job.
    """

    def on_close(event):
        print("Execution stopped by user!")
        current_job.halt()
        event.canvas.stop_event_loop()

    figure.canvas.mpl_connect("close_event", on_close)



from typing import List

import numpy as np

from .command_registry import *


class SequenceTracker:
    """
    Tracks a randomly-generated sequence by recording both a list of qubit gates
    corresponding to the sequence, and the raw command IDs which are used as
    input to the input stream to map into baked pulses.
    """

    def __init__(self, command_registry: CommandRegistry):
        self.command_registry: CommandRegistry = command_registry
        self._sequences_as_gates: list[Command] = []
        self._sequences_as_command_ids: list[list[int]] = []

    def make_sequence(self, command_ids: list[int]):
        """
        Adds a new sequence to the tracker, recorded as both a list
        of gates and command ids
        """
        self._record_sequence_as_list_of_command_ids(command_ids)
        self._record_sequence_as_list_of_gates(command_ids)

    def _record_sequence_as_list_of_command_ids(self, command_ids: list[int]):
        self._sequences_as_command_ids.append([])
        for command_id in command_ids:
            self._sequences_as_command_ids[-1].append(command_id)

    def _record_sequence_as_list_of_gates(self, command_ids: list[int]):
        self._sequences_as_gates.append([])
        for command_id in command_ids:
            command = self.command_registry.get_command_by_id(command_id)
            for gate in command:
                self._sequences_as_gates[-1].append(gate)

    def _serialize_sequences(self):
        result = ""
        for i, sequence in enumerate(self._sequences_as_gates):
            result += f"Sequence {i}:\n"
            result += f"\tCommand IDs: {self._sequences_as_command_ids[i]}\n"
            result += "\tGates:\n"
            for j, operation in enumerate(sequence):
                result += f"\t\t{j}: {operation}\n"
            result += "\n"
        return result

    def verify_sequences(self):
        """
        Checks that the application of all gates in a sequence to the |00>
        state correctly recovers to the |00> state at the end.
        """
        for i, sequence in enumerate(self._sequences_as_gates):
            ideal_state = self.calculate_resultant_state(sequence)
            expected_state = np.kron(np.array([1, 0]), np.array([1, 0]))  # should recover to |00>
            expected_state = np.outer(expected_state, expected_state.conj())
            assert np.allclose(
                ideal_state, expected_state
            ), f"expected to recover to at {expected_state}, got {ideal_state}"

        print(f"Verification passed for all {len(self._sequences_as_gates)} sequence(s).")

    def calculate_resultant_state(self, sequence: List[Gate]) -> np.ndarray:
        """
        Calculate the result of applying a list of gates in `sequence` to
        the multi-qubit ground-state.
        """
        ground_state = np.kron(np.array([1, 0]), np.array([1, 0]))
        ground_state_rho = np.outer(ground_state, ground_state.conj())
        rho = ground_state_rho
        for gate in sequence:
            rho = gate.matrix() @ rho @ gate.matrix().conj().T

        return rho

    def print_sequences(self):
        print(self._serialize_sequences())

    def save_to_file(self, path: Union[str, Path]):
        with open(path, "w+") as f:
            f.write(self._serialize_sequences())



from enum import Enum
from typing import Union, List

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from qualang_tools.wirer.connectivity.channel_spec import ChannelSpec
    from qualang_tools.wirer.connectivity.element import Element


class WiringFrequency(Enum):
    DC = "DC"
    RF = "RF"


DC = WiringFrequency.DC
RF = WiringFrequency.RF


class WiringIOType(Enum):
    INPUT = "input"
    OUTPUT = "output"
    INPUT_AND_OUTPUT = "input/output"


INPUT = WiringIOType.INPUT
OUTPUT = WiringIOType.OUTPUT
INPUT_AND_OUTPUT = WiringIOType.INPUT_AND_OUTPUT


class WiringLineType(Enum):
    RESONATOR = "rr"
    DRIVE = "xy"
    FLUX = "z"
    CHARGE = "q"
    COUPLER = "c"
    CROSS_RESONANCE = "cr"
    ZZ_DRIVE = "zz"


RESONATOR = WiringLineType.RESONATOR
DRIVE = WiringLineType.DRIVE
FLUX = WiringLineType.FLUX
CHARGE = WiringLineType.CHARGE
COUPLER = WiringLineType.COUPLER
CROSS_RESONANCE = WiringLineType.CROSS_RESONANCE
ZZ_DRIVE = WiringLineType.ZZ_DRIVE


class WiringSpec:
    """
    A technical specification for the wiring that will be required to
    manipulate the given quantum elements.
    """

    def __init__(
        self,
        frequency: WiringFrequency,
        io_type: WiringIOType,
        line_type: WiringLineType,
        triggered: bool,
        constraints: "ChannelSpec",
        elements: Union["Element", List["Element"]],
    ):
        self.frequency = frequency
        self.io_type = io_type
        self.line_type = line_type
        self.triggered = triggered
        self.constraints = constraints
        if not isinstance(elements, list):
            elements = [elements]
        self.elements: List["Element"] = elements
